/**
 * \file mgui/formbook.h
 * \brief MGUI::FORM_BOOK definition
 *
 * \if NODOC
 * $Id
 *
 * $Log: formbook.h_v $
 * Revision 1.2  2010/03/26 16:50:45  mju
 * Add ability to use ID for page.
 *
 * Revision 1.1  2010/03/26 16:36:43  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_MGUI_FORMBOOK_H
#define	INC_MGUI_FORMBOOK_H

#ifndef INC_MGUI_FORM_H
#include <mgui/form.h>
#endif

namespace MGUI {

//=====================================================================================================================

//! Form containing multiple "pages", only one of which is shown at at time.
//! All pages are the same size, which avoids issues with resizing when switching pages.
class FORM_BOOK : public FORM_COMPOSITE {
	public:

		//! Style for pages.
		enum PAGESTYLE {
			PAGESTYLE_Default =					0,
			PAGESTYLE_IgnoreNonVisible =		0x01,		//!< Ignore non-visible items for layout of page
			};

		//! Constructor.
		FORM_BOOK ();

		//! Destructor.
		virtual ~FORM_BOOK ();

		//! Add page to form, must be called after Create().
		//! @return Index of page, use GetPagePane() to access pane for creating controls on page.
		int AddPage (
			MGUI::LAYOUT_ORIENTATION orientation = MGUI::LAYOUT_ORIENTATION_Vertical,
			PAGESTYLE pagestyle = PAGESTYLE_Default,
			int ChildSpacing = 4,
			int ExtraBorder = 0
			);

		//! Create the form, must be called before AddPage().
		void Create (
			MGUI::LAYOUT_PANE_BASE& PaneParent,
			LAYOUT_SIZEALIGN sizealign = LAYOUT_SIZEALIGN_FixedSize,
			int ExtraBorder = 0
			);

		//! Get index to 'active' page.
		//! @return index to active page, -1 if none created.
		int GetActivePage () const
			{ return (m_ActivePage); }

		//! Get page ID.
		MGUI::ID GetPageID (
			int pageindex									//!< Page index as returned by AddPage()
			) const { return (m_PageIds.IsEmpty() ? 0 : m_PageIds[pageindex]); }

		//! Get reference to pane for specified page index.
		//! Do not do SetVisible() on returned LAYOUT_PANE, visibility must be changed via SetActivePage().
		LAYOUT_PANE& GetPagePane (
			int pageindex									//!< Page index as returned by AddPage()
			) { return (*m_PagePanePtrs[pageindex]); }

		//! Get number of pages.
		int GetNumPages () const
			{ return (m_PagePanePtrs.GetNumItems()); }

		//! Set active page by index.
		void SetActivePage (
			int pageindex									//!< Page index as returned by AddPage()
			);

		//! Set active page by ID.
		void SetActivePageID (
			MGUI::ID id
			);

		//! Set page ID.
		void SetPageID (
			int pageindex,									//!< Page index as returned by AddPage()
			MGUI::ID id
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		SIMPLE_ARRAY<LAYOUT_PANE*> m_PagePanePtrs;
		SIMPLE_ARRAY<MGUI::ID> m_PageIds;
		int m_ActivePage;

		FORM_BOOK (const FORM_BOOK&);
		FORM_BOOK& operator= (const FORM_BOOK&);
		#endif
	};

#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(FORM_BOOK::PAGESTYLE)
#endif

//=====================================================================================================================
}	// End namespace MGUI

#endif	// INC_MGUI_FORMBOOK_H
