/**
 * \file mgui/menubtn.h
 * \brief Definition for MGUI::CTRL_MENU_BUTTON.
 *
 * \if NODOC
 * $Id: menubtn.h_v 1.26 2011/11/09 15:51:24 mju Exp $
 *
 * $Log: menubtn.h_v $
 * Revision 1.26  2011/11/09 15:51:24  mju
 * Remove mfc stuff.
 *
 * Revision 1.1  2002/12/26 21:57:25  dwilliss
 * Initial revision
 * \endif
**/

#ifndef INC_MGUI_MENUBTN_H
#define INC_MGUI_MENUBTN_H

#ifndef	INC_MGUI_CTRL_H
#include <mgui/ctrl.h>
#endif

#ifndef	INC_MI32_MILIST_H
#include <mi32/milist.h>
#endif


namespace MGUI {

//! Control for an icon or text button that shows a menu.
class CTRL_MENUBUTTON : public MGUI::CTRL {
	public:

		//! Style flags for items.
		enum ITEMSTYLE {
			ITEMSTYLE_Normal			= 0,
			ITEMSTYLE_Toggle			= 0x00000001
			};

		enum STYLE {
			STYLE_Normal				= 0,
			STYLE_NoBorder				= 0x00000001,
			STYLE_HighlightOnEnter	= 0x00000002,
			STYLE_LargeButton			= 0x00000004
			};

		//! Constructor.
		CTRL_MENUBUTTON (
			) : m_menu(0), m_toppixel(0), m_bottompixel(0), m_bMenuMapped(false)
			{ }

		//! Destructor.
		virtual ~CTRL_MENUBUTTON (
			);

		//! Add menu item with Unicode label.
		//!
		//! The id will be passed to OnSelected() and must be unique within the
		//! dialog (for X it only has to be unique within the menu).  If you pass
		//! -1, an ID will be assigned.
		//!
		//! @return ID
		ID AddItem (
			ID id,
			const MISTRING& label,
			ITEMSTYLE style = ITEMSTYLE_Normal
			);

		//! Create control with text label.
		void Create (
			MGUI::LAYOUT_PANE_BASE& ParentPane,		//!< Parent pane
			const MISTRING& label,						//!< Label string
			STYLE style = STYLE_Normal,				//!< Button style
			MGUI::LAYOUT_SIZEALIGN sizealign = MGUI::LAYOUT_SIZEALIGN_FixedSize
			);

		//! Create control with icon and ToolTip.
		void Create (
			MGUI::LAYOUT_PANE_BASE& ParentPane,		//!< Parent pane
			ICONID iconid,									//!< Icon ID
			const MISTRING& tooltip,						//!< ToolTip string
			STYLE style = STYLE_Normal,				//!< Button style
			int iconsize = 0								//!< Icon size, 0 for default
			);

		//! Get the toggle state of a menu item.
		bool GetItemState (
			ID id
			) const;
		
		//! Get the number of items on the menu.
		int GetNumItems () const;

		//! Remove all items from menu.
		void RemoveAllItems ();

		//! Remove item by ID.
		void RemoveItemByID (
			ID id
			);

		//! Remove item by index.
		void RemoveItemByIndex (
			int indx
			);

		//! Set delegate to call when selection made from menu.
		void SetDelegateOnSelection (
			DELEGATE_VOID_ID delegate
			) { m_DelegateOnSelection = delegate; }

		//! Enable or disable an item
		void SetItemEnabled (
			ID id,
			bool enabled = true
			);

		//! Set the toggle state of a menu item.
		void SetItemState (
			ID id,
			bool state
			);

	protected:

		//! Called just before the menu is popped up.
		//! Default implementation does nothing.
		virtual void OnMenuPopup (
			);

		//! Called when a menu item is selected.
		//! Default implementation calls delegate.
		virtual void OnSelection (
			ID id									//!< ID of the item selected
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		struct ITEM {
			ID m_id;
			Widget m_widget;
			};

		DELEGATE_VOID_ID m_DelegateOnSelection;
		MILIST<ITEM>m_items;
		Widget m_menu;		// The popup menu
		MPIXEL m_toppixel;
		MPIXEL m_bottompixel;
		bool m_bMenuMapped;
		
		static void CB_Select (Widget, CTRL_MENUBUTTON*, void*);
		static void CB_MenuPopup (Widget, CTRL_MENUBUTTON*, void*);
		static void EH_Highlight (Widget, CTRL_MENUBUTTON*, void*, void*);

		bool CreateIcon (
			LAYOUT_PANE_BASE& ParentPane,
			ICONID iconid,
			const char *name,
			STYLE style,
			int iconsize
			);

		void CreateWidget (
			LAYOUT_PANE_BASE& ParentPane,
			XTARGS& args,
			const char *name,
			STYLE style
			);

		Widget GetItemWidget (
			ID id
			) const;

		CTRL_MENUBUTTON (const CTRL_MENUBUTTON&);
		CTRL_MENUBUTTON& operator= (const CTRL_MENUBUTTON&);
		#endif // GENERATING_DOXYGEN_OUTPUT
	};

#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(MGUI::CTRL_MENUBUTTON::ITEMSTYLE)
DEFINE_ENUM_OP_BITWISE(MGUI::CTRL_MENUBUTTON::STYLE)
#endif

} // end of namespace

#endif // INC_MGUI_MENUBTN_H
