/**
 * \file mgui/tooltip.h
 * \brief Definitions for MGUI::CTRL_TOOLTIP class
 *
 * \if NODOC
 * $Id: tooltip.h_v 1.20 2012/07/16 14:00:45 mju Exp $
 *
 * $Log: tooltip.h_v $
 * Revision 1.20  2012/07/16 14:00:45  mju
 * Add hideAll, remove setDelegateOnHide/Show as unused.
 *
 * Revision 1.19  2012/01/27 16:19:27  mju
 * Delegate for showrequest now global to mgui namespace.
 *
 * Revision 1.18  2012/01/27 16:04:37  mju
 * Add tooltip parm to delegate.
 *
 * Revision 1.17  2012/01/27 15:46:47  mju
 * Add delegates.
 *
 * Revision 1.16  2011/10/13 21:00:13  mju
 * Remove redundant hdr.
 *
 * Revision 1.15  2011/03/11 23:12:03  bcolombe
 * Add v_HasDatatip
 *
 * Revision 1.12  2009/10/15 16:28:22  mju
 * Add attach using ctrl.
 *
 * Revision 1.11  2007/05/21 17:27:20  dwilliss
 * Added parm to getimages method
 *
 * Revision 1.10  2007/05/21 15:16:29  dwilliss
 * added method to get images back out
 *
 * Revision 1.9  2007/01/26 16:42:56  dwilliss
 * Added SetAlpha method
 *
 * Revision 1.8  2007/01/26 15:38:06  dwilliss
 * Added AppendText and AppendImage methods
 *
 * Revision 1.6  2006/12/01 22:38:25  dwilliss
 * Added version of SetText which takes an IMAGE::STAGE_ARRAY
 *
 * Revision 1.4  2004/10/21 14:48:30  mju
 * Add hideNoTimer method.
 *
 * Revision 1.3  2004/08/25 17:55:44  mju
 * Add member to get text from widget.
 *
 * Revision 1.1  2004/02/05 22:06:44  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_MGUI_TOOLTIP_H
#define	INC_MGUI_TOOLTIP_H

#ifndef	INC_MGUI_BASE_H
#include <mgui/base.h>
#endif

#ifndef	INC_MI32_COLOR_H
#include <mi32/color.h>
#endif

#ifndef	INC_MI32_XBASE_H
#include <mi32/xbase.h>
#endif


#ifndef GENERATING_DOXYGEN_OUTPUT
// Forward declarations.
struct LPOINT2D;
class MISTRING;
namespace MGD {
	class DEVICE;
	class DEVICE_BINARY;
	}
namespace SPATMOD {
	namespace IMAGE {
		class STAGE;
		class STAGE_ARRAY;
		}
	}
#endif


namespace MGUI {

class CTRL_TOOLTIP {
	public:

		//! ToolTip style.
		struct STYLE {
			COLOR m_BackgroundColor;
			COLOR m_BorderColor;
			COLOR m_TextColor;
			INT16 m_MarginWidth;
			INT16 m_MarginHeight;
			// Default constructor.
			STYLE ();
			};

	#ifdef X_NATIVE
		//! Find CTRL_TOOLTIP instance attached to Widget (X only).
		//! @return NULL if none attached.
		static CTRL_TOOLTIP* FindAttached (
			Widget widget
			);
	#endif

		//! Hide all currently visible tooltips.
		static void HideAll ();

		//! Constructor.
		CTRL_TOOLTIP ();

		//! Destructor.
		virtual ~CTRL_TOOLTIP ();
		
		//! Append an image to the tooltip.
		void AppendImage (
			const MGD::DEVICE& device,
			bool bTakeOwnership = false
			);

		//! Append an image to the tooltip.
		void AppendImage (
			SPATMOD::IMAGE::STAGE& image
			);

		//! Append text to the tooltip.
		void AppendText (
			const MISTRING& text
			);

		//! Set text to show next time ToolTip is displayed.
		//! Does not cause ToolTip to appear.
		//! 
		//! The Images array is used if the text contains {~IMG #}, where the
		//!	# is a 0-based index into the Images array.  Note, even if the 
		//!   Tool tip already has images embeded in it, the the {~IMG #} tags
		//!	specified in the text parameter should be numbered starting at 0
		//!	relative to the Images array passed in.  They will be adjusted if
		//!	necessary.
		void AppendText (
			const MISTRING& text,							//!< Text to show
			const SPATMOD::IMAGE::STAGE_ARRAY* Images	//!< Images inserted into text.  
			);

		//! Attach instance to CTRL.
		void Attach (
			MGUI::CTRL& ctrl,									//!< Control to attach, must have already been created
			bool DeleteOnCtrlDestroy = false				//!< Delete ToolTip instance when associated control is destroyed
			);

	#ifdef X_NATIVE
		//! Attach instance to Widget (X only)
		void Attach (
			Widget widget,										//!< Widget to attach to
			bool DeleteOnCtrlDestroy = false				//!< Delete ToolTip instance when associated control is destroyed
			);
	#endif

		bool CheckMotion (
			int x,
			int y,
			bool AutoHide = true
			);

	#ifdef X_NATIVE
		bool CheckMotion (
			const XEvent& event,
			bool AutoHide = true
			);
	#endif

		//! Clear the current contents.
		//! Useful if you intend to Append to the tip.
		void Clear ();
		
		//! Detach instance from control.
		void Detach ();

		//! Disable automatic ToolTip display mechanism.
		//! ToolTip may still be shown by calling Show().  Does not hide ToolTip if presently showing.
		//! Each use of Disable() much be matched by call to Enable() to actually re-enable ToolTip
		void Disable ();

		//! Enable automatic ToolTip display mechanism.
		void Enable ();

		//! Get delay for automatic ToolTip display.
		//! @return Delay in milliseconds.
		INT32 GetDelay ();

		//! Get the images (if any) from the tooltip.
		//! The Images array is not cleared by this process.  The CTRL_TOOLTIP's
		//! images will be appended to the Images array.
		//! If bTransferOwnership is true, then the images will be removed from
		//! the CTRL_TOOLTIP and the Images array will "own" them.  If it's false,
		//! then it has to copy the images.
		ERRVALUE	GetImages (
			SPATMOD::IMAGE::STAGE_ARRAY& Images,
			bool bTransferOwnership
			);

		//! Get text to be displayed.
		//! @return Text string, will be blank if using label from attached control.
		const MISTRING& GetText () const;

		//! Get amount of movement for automatically hiding ToolTip.
		//! @return Movement amount in pixels.
		int GetPixelDelta () const;

		//! Get Priority for selecting amoung multiple ToolTips attached to same control.
		int GetPriority () const;

		//! Get current style settings.
		const STYLE& GetStyle () const;

		//! Hide ToolTip if visible.
		//! This is usually done automatically unless disabled.
		void Hide ();

		//! Determine if automatic display mechanism is disabled.
		//! Manual display via Show/Hide is always allowed.
		bool IsDisabled () const;

		//! Determine if ToolTip is presently visible.
		bool IsVisible () const;

		//! Set the Alpha (opacity) of the tooltip window.
		//! Note, this will only work under Windows, as it uses a feature unique
		//! to the MicroImages X Server.
		//! Alpha must be set before the tool tip is shown.
		void SetAlpha (
			int alpha		//!< 0 = transparent, 255 = opaque
			);

		//! Set delay for automatic ToolTip display.
		void SetDelay (
			INT32 delay											//!< ToolTip popup delay in milliseconds
			);

		//! Set delegate to call when requesting tooltip be shown.
		//! Delegate is only called if v_OnShowRequest is not overridden.
		void SetDelegateOnShowRequest (
			DELEGATE_ONTOOLTIPREQUEST delegate			//!< Delegate
			);

		//! Set image, optional mask and position for graphical ToolTip.
		//! This does not actually display the ToolTip.
		void SetImage (
			const MGD::DEVICE& imgdev,						//!< Device containing associated image
			const MGD::DEVICE_BINARY* maskdev = 0,		//!< Optional device containing mask
			const LPOINT2D* pos = 0							//!< Optional position relative to cursor
			);

		//! Set distance cursor must move to hide tooltip.
		//! Note that if cursor moves outside attached control the tooltip will be hidden
		//! regardless of distance.
		void SetPixelDelta (
			int PixelDelta										//!< Movement distance in screen pixels
			);

		//! Set priority if multiple CTRL_TOOLTIP instances are attached to the same control.
		//! If multiple instances attached the one with the highest priority will be shown.
		void SetPriority (
			int priority										//!< Priority to set
			);

		//! Set style to use next time ToolTip text is shown.
		void SetStyle (
			const STYLE& style
			);

		//! Set text to show next time ToolTip is displayed.
		//! Does not cause ToolTip to appear.
		void SetText (
			const MISTRING& text								//!< Text to show
			);

		//! Set text to show next time ToolTip is displayed.
		//! Does not cause ToolTip to appear.
		//! 
		//! The Images array is used if the text contains {~IMG #}, where the
		//!	# is a 0-based index into the Images array.
		void SetText (
			const MISTRING& text,							//!< Text to show
			const SPATMOD::IMAGE::STAGE_ARRAY* Images	//!< Images inserted into text.  
			);

		//! Show the ToolTip.
		//! This is usually done automatically unless disabled.
		void Show ();

	private:
		// Overridables.

		//! Override to set when datatip becomes available
		//! Used for making request in separate thread
		//! Default implementation returns 'true' unless no text or image is available.
		virtual bool v_HasDatatip ();

		//! Called after ToolTip is hidden.
		//! Generally used to send notifications about action.
		//! Default implementation does nothing.
		virtual void v_OnHide ();

		//! Called after ToolTip is shown.
		//! Generally used to send notifications about action.
		//! Default implementation does nothing.
		virtual void v_OnShow ();

		//! Override to perform custom ToolTip creating.
		//! When overriding, use SetText or SetImage to control what will be displayed.
		//! @return true to show ToolTip, false to not.
		//! Default implementation calls delegate if available otherwise returns 'true' unless no text or image is available.
		virtual bool v_OnShowRequest (
			int x,
			int y
			);

		#ifndef GENERATING_DOXYGEN_OUTPUT

		class PRIV;
		PRIV *m_pPriv;
		friend class PRIV;

		CTRL_TOOLTIP (const CTRL_TOOLTIP&);
		CTRL_TOOLTIP& operator= (const CTRL_TOOLTIP&);
		#endif // GENERATING_DOXYGEN_OUTPUT
	};

}	// End namespace MGUI

#endif	// INC_MGUI_TOOLTIP_H
