/**
 * \file mi32/cgm.h
 * \brief A class for reading CGM files.
 *
 * \if NODOC
 * $Log: cgm.h_v $
 * Revision 1.22  2011/10/21 20:50:07  mju
 * Remove redundant hdrs.
 *
 * Revision 1.21  2011/10/05 20:14:53  mju
 * Use trans2daffine hdr.
 *
 * Revision 1.20  2010/06/08 15:14:01  mju
 * Add m_transScale.
 *
 * Revision 1.18  2006/10/26 20:39:17  dwilliss
 * Added method for rendering PLANECONIC.  Got rid of seperate Arc and EArc
 * methods.  Renamed all virtuals to start with v_
 *
 * Revision 1.11  2002/12/20 18:33:50  dwilliss
 * Added FigureBegin/FigureEnd
 *
 * Revision 1.9  2001/07/25 16:25:24  dwilliss
 * Added a RenderText method
 *
 * Revision 1.1  2000/05/12 20:30:57  dwilliss
 * Initial revision
 * \endif
**/

#ifndef INC_MI32_CGM_H
#define INC_MI32_CGM_H

#ifndef INC_MI32_TRANS2DAFFINE_H
#include <mi32/trans2daffine.h>
#endif

#ifndef INC_MI32_ELEMSTYL_H
#include <mi32/elemstyl.h>
#endif

#ifndef INC_MI32_UCSTRING_H
#include <mi32/ucstring.h>
#endif

#ifndef	INC_MI32_SCRNUNIT_H
#include <mi32/scrnunit.h>
#endif

#ifndef	INC_MI32_POLYLINE_H
#include <mi32/polyline.h>
#endif


class CGM;
class CGMPARM;
class CGMELEM;
class PLANECONIC;
class FILEPATH;

#ifndef GENERATING_DOXYGEN_OUTPUT
//! A Parameter for a CGM (Computer Graphics Metafile) Element.
//!
//!	Note, the constructor is private.  You can't create one yourself.
//! CGMELEM::GetParm() will return a const pointer to one of these.
//!
class CGMPARM {
	public:

		enum PARMTYPE {	
			PARMTYPE_REAL = 0,
			PARMTYPE_SIGNED,
			PARMTYPE_UNSIGNED,
			PARMTYPE_STRING,
			PARMTYPE_COLOR,
			PARMTYPE_POINT
			};

		PARMTYPE GetType (
			) const {
			return (static_cast<PARMTYPE>(m_type));
			}

		double GetValueDouble (
			) const {
			return (*((double*)GetData()));
			}
		
		INT32 GetValueSigned (
			) const {
			return (*((INT32*)GetData()));
			}

		INT32 GetValueUnsigned (
			) const {
			return (*((UINT32*)GetData()));
			}
		
		const MIUNICODE* GetValueString (
			) const {
			return ((const MIUNICODE*)GetData());
			}
		
		const COLOR& GetValueColor (
			) const {
			return (*((const COLOR*)GetData()));
			}
		
		const DPOINT2D& GetValuePoint (
			) const {
			return (*((const DPOINT2D*)GetData()));
			}
		
		//! Get the size of the data for this parameter
		const int GetDataSize (
			) const {
			return m_size - sizeof(CGMPARM);
			}

		//! Get the size of data plus parameter header.
		//! If you cast a pointer to "this" to a (UINT8*)
		//! and add the size returned by this method, then cast
		//! it back to a (CGMPARM*), it will point to the next
		//! parameter.  (Don't do that, there are better ways,
		//! but GGM::ELEMENT::GetParm() uses this to do it.
		const int GetFullSize (
			) const {
			return m_size;
			}

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		CGMPARM() {}	//! Constructor is private.  Can't create one of these yourself!

		//! The data for a CGMPARM always imediately follows it in memory.
		const void* GetData (
			) const {
			return (this + 1);	
			}

		INT32 m_type;			//!< Could be an enum, but need to insure sizeof(CGMPARM) == 8
		UINT32 m_size;			//!< Size of this structure plus the parameter data that follows

		friend class CGM;
		#endif // GENERATING_DOXYGEN_OUTPUT
	};
#endif // GENERATING_DOXYGEN_OUTPUT


//! Computer Graphics Metafile reader
//!
//! This is meant to be a base class.  There are a number of protected
//! virtual methods for actually "rendering" the CGM object.
//!
//!	This class is NOT for writing CGM files.
//!
class CGM {
	public:
		
		enum ASF {	//! Aspect Source Flags
			ASF_INDIVIDUAL = 0,
			ASF_BUNDLED									 
			};

		enum VDCTYPE {
			VDCTYPE_INTEGER,
			VDCTYPE_REAL
			};

		enum REALMODE {
			REALMODE_FIXED = 0,
			REALMODE_FP
			};

		enum CLIPMODE {
			CLIPMODE_NONE,
			CLIPMODE_LOCUS,		//!< Hide whole element if centroid outside clip region
			CLIPMODE_SHAPE,		//!< Clip element to clip region
			CLIPMODE_LOCUSTHENSHAPE
			};

		enum COLORMODEL {			 //! XXX Need real values
			COLORMODEL_RGB,
			COLORMODEL_CMYK
			};

		enum COLORMODE {
			COLORMODE_INDEXED,
			COLORMODE_DIRECT
			};

		enum INTERIORSTYLE {
			INTERIORSTYLE_HOLLOW,
			INTERIORSTYLE_SOLID,
			INTERIORSTYLE_PATTERN,
			INTERIORSTYLE_HATCH,
			INTERIORSTYLE_EMPTY,
			INTERIORSTYLE_GEOMETRIC,
			INTERIORSTYLE_INTERPOLATED
			};

		enum LINETYPE {
			LINETYPE_SOLID = 1,
			LINETYPE_DASH,
			LINETYPE_DOT,				
			LINETYPE_DASH_DOT,
			LINETYPE_DASH_DOT_DOT
			};

		enum LINECAP {				//! XXX Need real values
			LINECAP_UNSPECIFIED=1,
			LINECAP_BUTT,
			LINECAP_ROUND,
			LINECAP_PROJECTING,
			LINECAP_TRIANGLE
			};

		enum LINEJOIN {			//! XXX Need real values
			LINEJOIN_UNSPECIFIED=1,
			LINEJOIN_MITRE,
			LINEJOIN_ROUND,
			LINEJOIN_BEVEL
			};

		enum LINECONT {
			LINECONT_UNSPECIFIED=1,
			LINECONT_CONTINUE,
			LINECONT_RESTART,
			LINECONT_ADAPTIVE
			};

		enum MARKERTYPE {
			MARKERTYPE_DOT=1,
			MARKERTYPE_PLUS,
			MARKERTYPE_ASTERISK,
			MARKERTYPE_CIRCLE,
			MARKERTYPE_EX
			//! Values above 5 are reserved for registration and future
			//! standardization
			};

		enum TEXTALIGNMENT {
			TEXTALIGNMENT_NORMAL,
			TEXTALIGNMENT_LEFT,		//!< For Horizontal alignment only
			TEXTALIGNMENT_CENTER,		//!< For Horizontal alignment only
			TEXTALIGNMENT_RIGHT,		//!< For Horizontal alignment only
			TEXTALIGNMENT_CONTINUOUS,
			TEXTALIGNMENT_TOP,		//!< For Vertical alignment only
			TEXTALIGNMENT_CAP,		//!< For vertical alignment only
			TEXTALIGNMENT_HALF,		//!< For vertical alignment only
			TEXTALIGNMENT_BASE,		//!< For vertical alignment only
			TEXTALIGNMENT_BOTTOM	//!< For vertical alignment only
			};

		enum TEXTPATH {
			TEXTPATH_LEFT,
			TEXTPATH_RIGHT,
			TEXTPATH_UP,
			TEXTPATH_DOWN
			};

		//! Indicates how text is clipped and how accurate placement is
		enum TEXTPRECISION {
			TEXTPRECISION_STRING,  //!< Ignores intracharacter spacing
			TEXTPRECISION_CHARACTER,
			STROKE
			};

		enum SCALEMODE {
			SCALEMODE_ABSTRACT = 0,
			SCALEMODE_METRIC
			};

		enum SIZEMODE {
			SIZEMODE_ABSOLUTE = 0,
			SIZEMODE_SCALED,
			SIZEMODE_FRACTIONAL,
			SIZEMODE_MM
			};

		enum VIEWPORTSPECMODE {
			VIEWPORTSPECMODE_FractionOfDrawingSurface,
			VIEWPORTSPECMODE_MMWithScaleFactor,
			VIEWPORTSPECMODE_DeviceCoordinates
			};


		// CONSTRUCTOR / DESTRUCTOR

		//! Constructor.
		//! The constructor only clears the structure.
		CGM (
			);

		virtual ~CGM (
			);

		// PUBLIC METHODS

		//! Return the VDC (Virtual Device Context?) extents.
		//! It should be noted that CGM files \e set this themselves.  It is not computed
		//! based on the elements in the file.  If the CGM file fails to set it, the default
		//! is 0 to 32767 in both directions.
		//! It should also be obvious that the extents are only valid after calling ReadFile().
		//! before that time, this function will return the default extents.
		void GetExtents (
			DRECT2D& rect
			) const {
			rect = m_VDCExtent;
			return;
			}

		//! Opens a CGM file and reads it into an internal buffer and parses it.
		int ReadFile (
			const FILEPATH& fname
			);

		//! Render the CGM data.
		//! The actual rendering is done by the protected virutal methods of a
		//! derived class.
		int Render (
			);

		//! Set the affine transformation.
		void SetTransAffine (
			const MAT3X3 forward,
			const MAT3X3 inverse
			);

		//! Set the affine transformation.
		void SetTransAffine (
			const TRANS2D_AFFINE& trans
			);


	protected:
		//!
		//!		The style information that doesn't eaisly fit into POINTSTYLE, etc...
		//!
		struct STYLEEXTRA {
			SCREENUNIT m_LineWidthUnits;			//!< Size units for line widths SCREENUNIT_Pixels, etc.
			SCREENUNIT m_EdgeWidthUnits;			//!< Size units for polygon edge widths (can be different than line widths)
			SCREENUNIT m_PointSizeUnits;			//!< Size units for points
			SCREENUNIT m_TextSizeUnits;			//!< Size units for Text

			STYLEEXTRA() :
				m_LineWidthUnits(SCREENUNIT_Points), 
				m_EdgeWidthUnits(SCREENUNIT_Points),
				m_PointSizeUnits(SCREENUNIT_Points),
				m_TextSizeUnits(SCREENUNIT_Points)
				{
				}
			};

		struct PRIMITIVECONTEXT {
			PRIMITIVECONTEXT();

			POINTSTYLE	m_PointStyle;
			LINESTYLE	m_LineStyle;
			POLYSTYLE	m_PolyStyle;
			TEXTSTYLE	m_TextStyle;
			STYLEEXTRA	m_StyleExtra;

			UINT32 m_id;							//!< Each context has a unique "name"
			UINT32  m_LineBundleIndex;				//!< Defualt == 1
			LINETYPE m_LineType;					//!< Default = (SOLID)
			double m_LineWidth;					//!< default == 1.0
			COLOR m_LineColor;			
			CLIPMODE m_LineClipMode;			//!< defualt == LOCUS
			LINECAP m_LineCap;					//!< default == UNSPECIFIED
			LINECAP m_DashCap;					//!< default == UNSPECIFIED
			LINEJOIN m_LineJoin;					//!< default == UNSPECIFIED
			LINECONT m_LineTypeContinuation;		//!< (1; unspecified)
			double m_LineTypeInitOffset;		//!< default = 0.0

			UINT32 m_MarkerBundleIndex;			//!< Default = 1
			MARKERTYPE m_MarkerType;			//!< Default = ASTERISK
			double m_MarkerSize;					//!< Default = 1.0
			COLOR m_MarkerColor;			
			CLIPMODE m_MarkerClipMode;			//!< defualt == LOCUS

			UINT32 m_TextBundleIndex;				//!< default = 1
			UINT32 m_TextFontIndex;					//!< default = 1
			TEXTPRECISION m_TextPrecision;	//!< default == STRING
			double m_CharExpFactor;				//!< default == 1.0
			double m_CharSpacing;				//!< default == 0.0
			COLOR m_TextColor;			
			double m_CharHeight;					//!< 1/100th of length of longest side
														//! of rectangle defined by default VDC
			double m_CharOrientXUp;				//!< Default= 0										
			double m_CharOrientYUp;				//!< Default=height of VDC
			double m_CharOrientXBase;			//!< Defaul=height of VDC
			double m_CharOrientYBase;			//!< Default=0
			TEXTPATH m_TextPath;					//!< default = RIGHT
			TEXTALIGNMENT m_TextHAlignment;	//!< default = NORMAL
			TEXTALIGNMENT m_TextVAlignment;	//!< default = NORMAL
			double m_TextContHAlign;			//!< default = 0.0
			double m_TextContVAlign;			//!< default = 0.0
			UINT32 m_CharSetIndex;					//!< default = 1
			UINT32 m_AltCharSetIndex;				//!< default = 1
			UINT32 m_TextScoreType;					//!< default == all off
			UINT32 m_RestrictedTextType;			//!< 1 (basic)

			UINT32 m_FillBundleIndex;				//!< default = 1
			INTERIORSTYLE m_InteriorStyle;	//!< default = HOLLOW
			COLOR m_FillColor;			
			UINT32 m_HatchIndex;						//!< Default = 1
			UINT32 m_PatternIndex;					//!< Default = 1

			UINT32 m_EdgeBundleIndex;				//!< Default = 1
			LINETYPE m_EdgeType;					//!< Default = SOLID
			double m_EdgeWidth;					//!< Default = 1.0
			COLOR m_EdgeColor;			
			bool m_bEdgeVisibility;				//!< default = false
			CLIPMODE m_EdgeClipMode;			//!< defualt == LOCUS
			LINECAP m_EdgeCap;					//!< default == UNSPECIFIED
			LINECAP m_EdgeDashCap;				//!< default == UNSPECIFIED
			LINEJOIN m_EdgeJoin;					//!< default == UNSPECIFIED
			LINECONT m_EdgeTypeContinuation;		//!< (1; unspecified)
			double m_EdgeTypeInitOffset;		//!< default = 0.0
			
			//! fill reference point
			DPOINT2D m_FillReferencePoint;
			//! pattern size
			//! Pick Identifier (for application use, ignore)
			//! symbol library index

			COLOR m_SymbolColor;
			double m_SymbolSize;
			double m_SymbolOrientXUp;				//!< Default= 0										
			double m_SymbolOrientYUp;				//!< Default=height of VDC
			double m_SymbolOrientXBase;			//!< Defaul=height of VDC
			double m_SymbolOrientYBase;			//!< Default=0

			//! clip indicator (default = on)
			//! clip rectangle (default = VDC Extent)
			DRECT2D m_ClipRect;
			bool m_bClipIndicator;
			UINT32 m_ProtectionRegionIndicator;	//!< 1; off?
			COLOR m_AuxColor;
			COLOR m_TransparentCellColor;		
			bool m_bTransparency;				//!< Default = false
			double m_MitreLimit;					//!< default = 32767.0;
			
			//! Aspect SourceFlags all default to INDIVIDUAL
			ASF m_asfLineType,
				 m_asfLineWidth,
				 m_asfLineColor,
				 m_asfMarkerType,
				 m_asfMarkerSize,
				 m_asfMarkerColor,
				 m_asfTextFontIndex,
				 m_asfTextPrecision,
				 m_asfCharExpFactor,
				 m_asfCharSpacing,
				 m_asfTextColor,
				 m_asfInteriroStyle,
				 m_asfFillColor,
				 m_asfHatchIndex,
				 m_asfPatternIndex,
				 m_asfEdgeType,
				 m_asfEdgeWidth,
				 m_asfEdgeColor;
			};

		// Protected methods

		//! Return the current Primitive Context
		const PRIMITIVECONTEXT& GetPC (
			) {
			return m_PC;
			}


		//! Draw a circular arc.
		//! If not implemented by the derived class, the default is to
		//! convert the conic to a line and call v_RenderPolyline or v_RenderPolygon
		virtual int v_RenderConic (
			const PLANECONIC& conic
			);

		//! Render a B-Spline.
		//!
		//!	If not overloaded by the derived class, the default method
		//! will convert the curve to points can call RenderPolyline()
		virtual int v_RenderBSpline (
			int order,
			int NumPts,
			const DPOINT2D* cpoints,		//!< Control points (NumPts of them)
			const double* knots,				//!< Array of (NumPts + order) knots
			const double* weights,			//!< Array of (NumPts weights) (may be NULL)
			double Start,
			double End
			);

		//! Render a Bezier curve.
		//!
		//!	If not overloaded by the derived class, the default method
		//! will convert the curve to points can call RenderPolyline()
		virtual int v_RenderBezier (
			int NumPts,
			DPOINT2D* cpoints,
			bool bContinuous
			);

		//! Draw a circle.
		//! If not implemented by the derived class, the default is to
		//! call v_RenderConic
		virtual int v_RenderCircle (
			const DPOINT2D& center,
			double radius
			);


		//! Render an Ellipse (Simple Case).
		//! If not implemented by the derived class, the default is to
		//! call v_RenderConic unless there's a lot of shear, in which case it will
		//! convert it to polygon and call v_RenderPolygon
		virtual int v_RenderEllipse (
			const DPOINT2D& center,
			double xradius,
			double yradius,
			double RotAngle
			);


		//! Render an Elliptical Arc (General Case).
		//!
		//!	This method takes exactly what CGM gives us.  Instead of 
		//! two radii and a rotation angle, they give the endpoints
		//! of where the axes intersect the ellipse.  Note This gives
		//! the possibility of shear angle too.  Ick!
		//!
		//! Override this method only if you are writing to a format
		//! that wants this weird format.  The default method will
		//! call either the other RenderEllipse (for the non-shear
		//! case or will convert the ellipse to a polygon and call
		//! RenderPolygon
		virtual int v_RenderEllipse (
			const DPOINT2D& center,
			const DPOINT2D& EndPoint1,
			const DPOINT2D& EndPoint2
			);

		//! Draw a 2-point line.
		//!
		//!	If not overloaded by the derived class, the default method just
		//! calls RenderPolyline with a 2-point line.
		virtual int v_RenderLine (
			const DPOINT2D& start,
			const DPOINT2D& end
			);

		//! Draw a marker (point).
		virtual int v_RenderMarker (
			const DPOINT2D& point
			);

		//! Draw a closed polygon.
		//!
		//! If not overloaded by the derived class, the default method just
		//! calls RenderPolygons() with one polygon.
		virtual int v_RenderPolygon (
			const POLYLINE& polygon
			);

		//! Render multiple polygons (with islands).
		virtual int v_RenderPolygons (
			const POLYLINELIST& polygons
			);

		//! Draw a multi-point line.
		virtual int v_RenderPolyline (
			const POLYLINE& line
			);

		//! Draw a rectangle.
		//!
		//!	Points will be in the order...
		//!
		//!		0     1
		//!
		//!		3		2
		//!
		//!	For convinence, you will actually be passed 5 points,
		//! and points[4] == points[0].
		//! If not overloaded by the derived class, the default method
		//! will call RenderPolygon.
		virtual int v_RenderRectangle (
			const DPOINT2D* points
			);

		//! Draw a rectangle (no rotation).
		//!
		//! If the current affine transformation has no rotation, this
		//! method will be called to render rectangles, since it can be
		//! much faster.  If not overloaded, the default method will 
		//! call the other RenderRectangle() method, so implementing this
		//! one is optional.
		virtual int v_RenderRectangle (
			const DRECT2D& rect
			);

		//! Render text.
		virtual int v_RenderText (
			const DPOINT2D& basept,
			const TEXTSTYLE& style,
			const STYLEEXTRA& extra,
			const MIUNICODE* str
			);

		//! Set point style.
		//!
		//!	Called whenever something about the current point style changes.  
		virtual int v_SetPointStyle (
			const POINTSTYLE& style,
			const STYLEEXTRA& extra
			);

		//! Set line style.
		//!
		//!	Called whenever something about the current line style changes.  
		virtual int v_SetLineStyle (
			const LINESTYLE& style,
			const STYLEEXTRA& extra
			);

		//! Set polygon style.
		//!
		//!	Called whenever something about the current polygon style changes.  
		virtual int v_SetPolyStyle (
			const POLYSTYLE& style,
			const STYLEEXTRA& extra
			);

		//! Set text style.
		//!
		//!	Called whenever something about the current text style changes.  
		virtual int v_SetTextStyle (
			const TEXTSTYLE& style,
			const STYLEEXTRA& extra
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		//! A Single CGM (Computer Graphics Metafile) element
		class ELEMENT {
			public:

			//! Return the CGM Class code (range 1 - 8).
			int GetClass (
				) const {
				return m_ElemClass;
				}

			//! Return the CGM ElementID.  
			//! The range depends on the CGM Class
			int GetElementID (
				) const {
				return m_ElemID;
				}

			//! Return the number of parameters for this element.
			int GetNumParms (
				) const {
				return m_NumParms;
				}

			//! Return the number of parameters for this element.
			//! Returns NULL if no such parameter
			const CGMPARM* GetParm (
				const CGM* cgm,
				int ParmNum
				) const;

			private:
				int m_ElemClass;
				int m_ElemID;
				int m_FirstParmOffset;
				int m_NumParms;

				friend class CGM;
			};	//! End of embeded class ELEMENT



		// PRIVATE METHODS

		int AddParm (
			ELEMENT* elem,
			double val
			);

		int AddParm (
			ELEMENT* elem,
			INT32 val
			);

		int AddParm (
			ELEMENT* elem,
			UINT32 val
			);

		int AddParm (
			ELEMENT* elem,
			const MIUNICODE* val
			);

		int AddParm (
			ELEMENT* elem,
			COLOR& val
			);

		int AddParm (
			ELEMENT* elem,
			DPOINT2D& val
			);

		void FigureBegin();
		void FigureEnd();
		void FigureFree();
		void FigureNewPoly();
		void FigureAddPoints (
			const POLYLINE& points
			);

		//! Return the angle represented by two parts of a unit vector.
		//! The vector componants are read as two parameters of the element.
		double GetAngle (
			CGMELEM& elem,
			int dxParmNum,
			int dyParmNum
			);

		///////////////////////////////////////////////////////////////////////////!
		#ifndef GENERATING_DOXYGEN_OUTPUT
		//! Methods to get POD data values  
		//! No, these are not sorted alphabetically, and if you sort them, I will
		//! have to hurt you!
		#endif //!< GENERATING_DOXYGEN_OUTPUT

		//! Get an 8-bit signed integer (SI8 in the CGM documentation)
		int GetSI8() {
			return *((signed char*)m_pNextByte++);
			}

		//! Get a 16-bit signed integer (SI16 in the CGM documentation)
		int GetSI16() {
			//! Yes, the following is unsigned even though we're returning signed.
			//! This keeps the compiler from second-guessing us. 
			UINT16 word = (m_pNextByte[0] << 8) | m_pNextByte[1];
			m_pNextByte += 2;
			return ((INT16)word);
			}

		//! Get a 24-bit signed integer (SI24 in the CGM documentation)
		int GetSI24() {
			//! Yes, the following is unsigned even though we're returning signed.
			//! This keeps the compiler from second-guessing us.  
			UINT32 word = (m_pNextByte[0] << 16) | (m_pNextByte[1] << 8) | m_pNextByte[2];
			if (word & 0x00800000) word |= 0xFF000000;	//!< Extend sign bit
			m_pNextByte += 3;								
			return ((INT32)word);
			}

		//! Get a 16-bit signed integer (SI32 in the CGM documentation)
		int GetSI32() {
			//! Yes, the following is unsigned even though we're returning signed.
			//! This keeps the compiler from second-guessing us. 
			UINT32 word = (m_pNextByte[0] << 24) | (m_pNextByte[1] << 16) | (m_pNextByte[2] << 8) | m_pNextByte[3];
			m_pNextByte += 4;								
			return ((INT32)word);
			}

		//! Get a signed integer at a given precision
		int GetSI (
			int precision
			) {
			switch (precision) {
				case 8: return (GetSI8());
				case 16: return (GetSI16());
				case 24: return (GetSI24());
				case 32: return (GetSI32());
				}
			return (0);	//! "can't" get here
			}

		//! Get an 8-bit unsigned integer (UI8 in the CGM documentation)
		UINT8 GetUI8() {
			return *m_pNextByte++;
			}

		//! Get an 8-bit unsigned Character (C in the CGM documentation)
		UINT8 GetC() {
			return *m_pNextByte++;
			}

		//! Get a 16-bit unsigned integer (UI16 in the CGM documentation)
		UINT16 GetUI16() {
			UINT16 word = (m_pNextByte[0] << 8) | m_pNextByte[1];
			m_pNextByte += 2;
			return ((INT16)word);
			}

		//! Get a 24-bit unsigned integer (UI24 in the CGM documentation)
		UINT32 GetUI24() {
			UINT32 word = (m_pNextByte[0] << 16) | (m_pNextByte[1] << 8) | m_pNextByte[2];
			m_pNextByte += 3;								
			return (word);
			}

		//! Get a 16-bit unsigned integer (UI32 in the CGM documentation)
		UINT32 GetUI32() {
			UINT32 word = (m_pNextByte[0] << 24) | (m_pNextByte[1] << 16) | (m_pNextByte[2] << 8) | m_pNextByte[3];
			m_pNextByte += 4;								
			return (word);
			}

		//! Get a signed integer at a given precision
		UINT32 GetUI (
			int precision
			) {
			switch (precision) {
				case 8: return (GetUI8());
				case 16: return (GetUI16());
				case 24: return (GetUI24());
				case 32: return (GetUI32());
				}
			return (0);	//! "can't" get here
			}

		//! Get a 32-bit Fixed-Point number (FX32 in the CGM documentation)
		double GetFX32() {
			double WholePart = (double)GetSI16();
			double FracPart = (double)GetUI16();
			return (WholePart + (FracPart / (double)(0xffffUL)));	//! 0xFFFF = 2^16
			}

		//! Get a 64-bit Fixed-Point number (FX64 in the CGM documentation)
		double GetFX64() {
			double WholePart = (double)GetSI32();
			double FracPart = (double)GetUI32();
			return (WholePart + (FracPart / (double)(0xffffffffUL)));	//! 0xFFFFFFFF = 2^32
			}

		//! Get a fixed point number at a given precision
		double GetFX (
			int precision
			) {
			switch (precision) {
				case 32: return (GetFX32());
				case 64: return (GetFX64());
				}
			return (0.0);	//! "can't" get here
			}

		//! Get a 32-bit Floating-Point number (FP32 in the CGM documentation)
		double GetFP32() {
			#ifdef BYTEORDER_LoHi
				UINT8 bytes[4];
				bytes[0] = *m_pNextByte++;
				bytes[1] = *m_pNextByte++;
				bytes[2] = *m_pNextByte++;
				bytes[3] = *m_pNextByte++;
				SwapBytes4(bytes, 1);
				return *(float*)bytes;
			#else
				float fp = *(float*)m_pNextByte;
				m_pNextByte += 4;
				return(fp);
			#endif
			}

		//! Get a 64-bit Floating-Point number (FP64 in the CGM documentation)
		double GetFP64() {
			#ifdef BYTEORDER_LoHi
				UINT8 bytes[8];
				bytes[0] = *m_pNextByte++;
				bytes[1] = *m_pNextByte++;
				bytes[2] = *m_pNextByte++;
				bytes[3] = *m_pNextByte++;
				bytes[4] = *m_pNextByte++;
				bytes[5] = *m_pNextByte++;
				bytes[6] = *m_pNextByte++;
				bytes[7] = *m_pNextByte++;
				SwapBytes8(bytes, 1);
				return *(double*)bytes;
			#else
				double fp = *(double*)m_pNextByte;
				m_pNextByte += 4;
				return(fp);
			#endif
			}

		//! Get a floating point at a given precision
		double GetFP (
			int precision
			) {
			switch (precision) {
				case 32: return (GetFP32());
				case 64: return (GetFP64());
				}
			return (0.0);	//! "can't" get here
			}

		///////////////////////////////////////////////////////////////////////////!
		//! Methods to get Abstract data values  

		double GetR (
			int precision,
			REALMODE mode
			) {
			if (mode == REALMODE_FIXED) {
				return (GetFX(precision));
				}
			else {
				return (GetFP(precision));
				}
			}

		double GetVDC() {
			if (m_VDCType == VDCTYPE_REAL) {
				return (GetR(m_vrp, m_VDCRealMode));
				}
			else {
				return (GetSI(m_vip));
				}
			}

		double GetVC() {
			if (m_VPSpecMode == VIEWPORTSPECMODE_FractionOfDrawingSurface) {
				return (GetR(m_rp, m_RealMode));
				}
			else {
				return (GetSI(m_ip));
				}
			}

		UINT32 GetLength () {
			UINT8 byte = GetUI8();
			if (byte < 255) return(byte);
			else return(GetLengthLongForm());
			}

		//! CGM encodes length parameters oddly.  The length for an element header
		//!	is encoded as the low 5 bits of the first word.  If the low 5 bits are
		//! 31, the actual length is too big to fit in 5 bits and follows as 2 or
		//! more 16-bit words.  This method will read that encoding.  Strings record
		//! their length the same way, but get a whole byte instead of just 5-bits
		//! to encode the short form.  If the byte is 255, the length is in this
		//! form.
		UINT32 GetLengthLongForm () {
			UINT32 length = 0;
			for (;;) {
				UINT16 word = GetUI16();
				length |= word & 0x7FFF;
				if (!(word & 0x8000)) return(length);
				length <<= 15;
				}
			}

		int ParmAlloc (
			CGMPARM::PARMTYPE type,
			const void* data,
			int size					//!< Size of parameter data
			);

		int ProcessClass0Element (CGMELEM& elem, bool bRender);
		int ProcessClass1Element (CGMELEM& elem, bool bRender);
		int ProcessClass2Element (CGMELEM& elem, bool bRender);
		int ProcessClass3Element (CGMELEM& elem, bool bRender);
		int ProcessClass4Element (CGMELEM& elem, bool bRender);
		int ProcessClass5Element (CGMELEM& elem, bool bRender);
		int ProcessClass6Element (CGMELEM& elem, bool bRender);
		int ProcessClass7Element (CGMELEM& elem, bool bRender);
		int ProcessClass8Element (CGMELEM& elem, bool bRender);

		int ProcessElement (
			ELEMENT* pelem,
			bool bRender
			);

		//! Parses the input file into in internal representation which can
		//! then be rendered.
		int ScanCGM ();

		int ScanCGMBinary ();
		int ScanCGMCharEncoded ();
		int ScanCGMClearText ();

		int ScanElemParms (
			ELEMENT* elem,
			const char* constParmSpec,
			int ParmListLen
			);

		PRIMITIVECONTEXT* GetPC (
			UINT32 id,
			bool bAllocIfMissing = false
			);

		// PRIVATE DATA

		SIMPLE_ARRAY<ELEMENT> m_elem;
		UINT8* m_buf;
		UINT8* m_pNextByte;		//!< Used when scanning the file
		int m_BufSize;
		int m_ParmBufNext;		//!< An offset into m_pParmBuf
		int m_ParmBufSize;
		UINT8* m_pParmBuf;

		// CGM states
		VDCTYPE m_VDCType;
		REALMODE m_RealMode;
		REALMODE m_VDCRealMode;
		SCALEMODE m_ScaleMode;
		double m_ScaleFactor;
		VIEWPORTSPECMODE m_VPSpecMode;
		double m_VPScaleFactor;
		UINT32 m_rp;				//!< Real precision
		UINT32 m_ip;				//!< Integer precision
		UINT32 m_ixp;				//!< Index precision
		UINT32 m_cip;				//!< Color Index Precision
		UINT32 m_dcp;				//!< Direct Color Precision
		UINT32 m_vrp;
		UINT32 m_vip;
		UINT32 m_namep;			//!< Name Precision
		double m_TransScale;
		double m_TransRotAngle;
		double m_TransShearAngle;
		MIUNICODE** m_FontList;
		MIUNICODE** m_CharSetList;
		//! Character Coding announcer (default = basic 7-bit)
		//! Maximum VDC Extent defaults to default VDC Extent
		
		COLORMODEL m_ColorModel;
		COLORMODE m_ColorMode;
		INT32 m_ColorCalib;	//!< 1 == unspecified
		SIZEMODE m_LineWidthMode;
		SIZEMODE m_MarkerSizeMode;
		SIZEMODE m_EdgeWidthMode;
		DRECT2D m_VDCMaxExtent;			//!< (0,32767, 0,32767), (0,1,0,1) if vdptype real
		DRECT2D m_VDCExtent;			//!< (0,32767, 0,32767), (0,1,0,1) if vdptype real
		DRECT2D m_DeviceViewPort;	//!< (0.0, 1.0, 0.0, 1.0)

		COLOR m_BGColor;
		LPOINT3D m_RGBMin;
		LPOINT3D m_RGBMax;
		DPOINT3D m_CIEScale;
		DPOINT3D m_CIEOffset;

		int m_NumPCs;
		PRIMITIVECONTEXT **m_PCList;
		PRIMITIVECONTEXT m_PC;
		POLYLINELIST m_FigurePolygons;
		bool m_bDefineFigure;

		//! Other things I don't have members for yet
		//! Device viewport specification mode (default = fraction of display)
		//! device viewport mapping (default = forced, left, bottom)
		//! line representation (interpreter dependent)
		//! marker representation (interpreter dependent)
		//! text representation (interpreter dependent)
		//! fill representation (interpreter dependent)
		//! edge representation (interpreter dependent)
		//! interior style specification mode (default = absolute)
		//! geometric pattern definition (default = 1; default VDC extent)
		//! Protection region region identical to CLIP rectangle
		//! TextPathMode;	(defult == off)
		//! Compound text path (default=line from lower-left to lower-right of
		//! 	default VDC extent)
		
		//! Pattern Table;
		COLOR *m_pColorTable;
		UINT32 m_MaxColorIndex;			//!< Default = 63

		TRANS2D_AFFINE m_trans;		//!< Affine projection

		MAT3X3 m_FwdAffine, m_InvAffine;

		friend class CGMELEM;	//!< Needs access to private embeded class ELEMENT
		friend class ELEMENT;
		#endif // GENERATING_DOXYGEN_OUTPUT
	};



#ifndef GENERATING_DOXYGEN_OUTPUT
//! A Single CGM (Computer Graphics Metafile) element.
//!
//!	Note, the constructor for this class is private.
//! The protected virtual methods of CGM (or one of its
//! derived classes) will pass you a const reference 
//! to a CGMELEM
//!
//! Note, this class is a wrapper around the internal representation of an Element.
//! The internal representation does not store a pointer to the parent CGM class
//! which we need for the GetParm() method.  I could have made that method take
//! a pointer to the CGM, but I thought that would look ugly.
class CGMELEM {
public:

	//! Return the CGM Class code (range 1 - 8).
	int GetClass (
		) const {
		return m_Elem.GetClass();
		}

	//! Return the CGM ElementID.  
	//! The range depends on the CGM Class
	int GetElementID (
		) const {
		return m_Elem.GetElementID();
		}

	//! Return the number of parameters for this element.
	int GetNumParms (
		) const {
		return m_Elem.GetNumParms();
		}

	//! Return the Nth parameter for this element.
	//! Returns NULL if no such parameter
	const CGMPARM* GetParm (
		int ParmNum
		) const {
		return m_Elem.GetParm(m_pCGM, ParmNum);
		}

	//! Return the Nth parameter as a double.
	//! Returns true if there was such a parameter, false if not
	bool GetParm (
		int ParmNum,
		double& retval
		) {
		const CGMPARM* parm = GetParm(ParmNum);
		if (!parm) return(false);
		retval = parm->GetValueDouble();
		return(true);
		}

	//! Return the Nth parameter as an unsigned integer.
	//! Returns true if there was such a parameter, false if not
	bool GetParm (
		int ParmNum,
		UINT32& retval
		) {
		const CGMPARM* parm = GetParm(ParmNum);
		if (!parm) return(false);
		retval = parm->GetValueUnsigned();
		return(true);
		}

	//! Return the Nth parameter as a signed integer.
	//! Returns true if there was such a parameter, false if not
	bool GetParm (
		int ParmNum,
		INT32& retval
		) {
		const CGMPARM* parm = GetParm(ParmNum);
		if (!parm) return(false);
		retval = parm->GetValueSigned();
		return(true);
		}

	//! Return the Nth parameter as a COLOR.
	//! Returns true if there was such a parameter, false if not
	bool GetParm (
		int ParmNum,
		COLOR& retval
		) {
		const CGMPARM* parm = GetParm(ParmNum);
		if (!parm) return(false);
		retval = parm->GetValueColor();
		return(true);
		}

	//! Return the Nth parameter as a DPOINT2D.
	//! Returns true if there was such a parameter, false if not
	bool GetParm (
		int ParmNum,
		DPOINT2D& retval
		) {
		const CGMPARM* parm = GetParm(ParmNum);
		if (!parm) return(false);
		retval = parm->GetValuePoint();
		return(true);
		}

	//! Return the Nth parameter as a (MIUNICODE*).
	//! Returns true if there was such a parameter, false if not.
	//! Caller must free the resulting string with MmFree()
	bool GetParm (
		int ParmNum,
		MIUNICODE** retval
		) {
		const CGMPARM* parm = GetParm(ParmNum);
		if (!parm || !retval) return(false);
		*retval = ucstrdup(parm->GetValueString());
		return(true);
		}

private:
	#ifndef GENERATING_DOXYGEN_OUTPUT
	CGMELEM (
		CGM* cgm,
		const CGM::ELEMENT& elem
		) :
		m_pCGM(cgm),
		m_Elem(elem)
		{ 
		}

	CGM* m_pCGM;
	const CGM::ELEMENT& m_Elem;

	friend class CGM;		//!< CGM can call our constructor
	#endif // GENERATING_DOXYGEN_OUTPUT
	};


#endif //!< GENERATING_DOXYGEN_OUTPUT

#endif  // INC_MI32_CGM_H
