/**
 * \file mi32/mg2glx.h
 * \brief MGD::DEVICE_GLX definitions.
 *
 * \if NODOC
 * $Id: mg2glx.h_v 1.6 2012/09/14 15:00:37 mju Exp $
 *
 * $Log: mg2glx.h_v $
 * Revision 1.6  2012/09/14 15:00:37  mju
 * Fix dox.
 *
 * Revision 1.5  2012/03/22 20:52:43  mju
 * Add v_getImagePointer.
 *
 * Revision 1.4  2012/03/13 20:09:19  mju
 * Make notifyChangeWindow pubic.
 *
 * Revision 1.3  2012/03/07 22:31:22  mju
 * Fix inclusion guard name.
 *
 * Revision 1.2  2012/03/07 22:14:41  mju
 * Add buffer for non-stereo.
 *
 * Revision 1.1  2012/03/07 17:30:33  mju
 * Initial revision
 * \endif
**/

#ifndef INC_MI32_MG2GLX_H
#define INC_MI32_MG2GLX_H

#ifndef INC_MI32_MGD2_H
#include <mi32/mgd2.h>
#endif

#ifndef INC_MI32_MIGLX_H
#include <mi32/miglx.h>
#endif


namespace MGD {

//! Device for drawing using OpenGL via GLX extension.
class DEVICE_GLX : public DEVICE {

	public:

		//! Enumeration for which OpenGL buffer(s) to associate device with.
		enum BUFFER {
			BUFFER_Front =					1,	//!< Front, with separate stereo frame buffers if available
			BUFFER_Back =					2,	//!< Back, with separate stereo frame buffers if available
			BUFFER_FrontAndBack =		3,	//!< Both front and back, with separate stereo frame buffers if available
			BUFFER_FrontLeftAndRight =	4,	//!< Front, drawing to both stereo buffers if available
			BUFFER_BackLeftAndRight =	5,	//!< Back, drawing to both stereo buffers if available
			BUFFER_All =					6	//!< All available buffers
			};

		//! Default constructor.
		DEVICE_GLX ();

		//! Destructor
		virtual ~DEVICE_GLX ();

		//! Create device associated with GLX context.
		//! Note that for stereo contexts, unless BUFFER_...LeftAndRight or All is specified, the device
		//! will report that 2 frame buffers are available and each may be rendered into separately.
		ERRVALUE Create (
			MIGLX::CONTEXT& GlxContext,	//!< Context to associate with, must remain valid for life of device
			XID xwindow,						//!< X window
			BUFFER buffer						//!< OpenGL buffer(s) to draw into
			);

		//! Determine if is stereo.
		bool IsStereo () const
			{ return (m_IsStereo); }

		//! Notify device that window (typically size) has changed.
		void NotifyChangeWindow ();

	private:
   	#ifndef GENERATING_DOXYGEN_OUTPUT
		typedef DEVICE BASECLASS;

		class CONTEXT;

		MIGLX::CONTEXT *m_pGlxContext;
		SIMPLE_ARRAY<UINT8> m_RowBuffer;
		XID m_xwindow;
		UINT32 m_glBuffer;
		UINT32 m_glBufferLeft;
		UINT32 m_glBufferRight;
		bool m_IsStereo;

		// Unimplemented
		DEVICE_GLX (const DEVICE_GLX&);
		DEVICE_GLX& operator= (const DEVICE_GLX&);

		void SetupGlxContext (UINT16 FrameBuffer);

		// Overrides.
		virtual bool v_AvoidFillSpans () const;
		virtual ERRVALUE v_Clear (const COLOR& color);
		virtual MGD::CONTEXT* v_CreateContext ();
		virtual ERRVALUE v_FillSpan (MGD::CONTEXT *gc, INT32 x1, INT32 x2, INT32 y);
		virtual CAPABILITY v_GetCapabilities () const;
		virtual UINT8* v_GetImagePointer (INT32 x, INT32 y, INT32 len, UINT16 FrameBuffer);
		virtual UINT16 v_GetNumFrameBuffers () const;
		#endif // GENERATING_DOXYGEN_OUTPUT

	};

} // End of namespace MGD


#endif	// INC_MI32_MG2GLX_H

