/******************************************************************************
 *
 * \file mi32/mg2rast.h
 * \brief A base raster-based MGD::DEVICE.  
 *
 * Currently there is only a 24-bit version.  The 16 bit version still
 * needs to be written, but should not be difficult
 *
 * This actually derives from MGD::DEVICE_MEM24.  The raster reading and
 * writing is handled by overloading GetImagePointer() and a couple other
 * methods.
 *
 * \if NODOC
 * $Log: mg2rast.h_v $
 * Revision 1.21  2012/02/20 18:40:33  mju
 * Change base classes.
 *
 * Revision 1.20  2011/10/26 17:58:11  mju
 * Frame buffer support.
 *
 * Revision 1.19  2011/03/09 15:02:12  mju
 * Use mem32rgba as base for raster_rgb32.
 *
 * Revision 1.18  2008/11/20 17:31:16  mju
 * Use getObjItem instead of getObjectItem.
 *
 * Revision 1.17  2006/09/20 15:43:17  dwilliss
 * Eliminate legacy method
 *
 * Revision 1.16  2006/04/13 16:24:43  mju
 * Add 32-bit device.
 *
 * Revision 1.15  2005/12/21 18:46:52  dwilliss
 * Changed to use interface classes
 *
 * Revision 1.14  2005/01/03 20:10:44  dwilliss
 * Moved SetVirtualWindow method to DEVICE
 *
 * Revision 1.13  2003/12/15 23:37:17  dwilliss
 * Diverging for 7.0
 * Made changes necessary for graphical data tips
 *
 * Revision 1.12  2003/12/10 23:47:45  dwilliss
 * Got rid of virtual window members. In the base class now.
 *
 * Revision 1.11  2003/10/01 22:16:42  dwilliss
 * Doxygen
 *
 * Revision 1.10  2003/09/15 13:49:56  fileserver!dwilliss
 * Doxygen
 *
 * Revision 1.9  2003/06/11 17:12:31  dwilliss
 * Virtual methods all renamed to start with v_ and made private (where
 * possible) or protected.
 *
 * Revision 1.8  2002/11/18 16:11:43  mju
 * Add bgcolor parm to Create().
 *
 * Revision 1.7  2002/02/06 16:11:59  dwilliss
 * Added method to set virtual window
 *
 * Revision 1.6  2002/01/29 17:07:00  dwilliss
 * Added 16-bit raster device
 *
 * Revision 1.5  2002/01/09 16:58:00  dwilliss
 * FillSpanDirect needed const parms
 *
 * Revision 1.1  2001/07/11 17:33:48  dwilliss
 * Initial revision
 *
 * \endif
 */

#ifndef INC_MI32_MG2RAST_H
#define INC_MI32_MG2RAST_H

#ifndef INC_MI32_MG2MEM16_H
#include <mi32/mg2mem16.h>
#endif

#ifndef INC_MI32_MG2MEM24_H
#include <mi32/mg2mem24.h>
#endif

#ifndef INC_MI32_MG2MEM32RGBA_H
#include <mi32/mg2mem32rgba.h>
#endif

#ifndef INC_RVC_RASTER_H
#include <rvc/raster.h>
#endif

namespace MGD {

//=====================================================================================================================

class CONTEXT_RASTER_RGB24;

//! Device for rendering to 24-bit RVC raster.
//! Raster must have CELLTYPE_RedGreenBlue_24 or BlueGreenRed_24.
class DEVICE_RASTER_RGB24: public DEVICE_MEM24 {

	public:

		//! Construct a 24-bit RGB raster device.
		//! After constructing, you should call Create()
		DEVICE_RASTER_RGB24();

		//! Destructor
		virtual ~DEVICE_RASTER_RGB24();

		// (Note: copy constructor private and unimplemented.  Can't copy one.)

		//! Create a raster to draw into.
		ERRVALUE Create (
			RVC::OBJECT& ParentObject,
			const RVC::DESCRIPTOR& Descriptor,
			RVC::RASTER::MAKEPARMS& MakeParms,
			const RVC::CREDENTIALS* Credentials = 0
			);

		//! Create a raster to draw into.
		ERRVALUE Create (
			RVC::OBJITEM& ObjItem,
			RVC::RASTER::MAKEPARMS& MakeParms,
			const RVC::CREDENTIALS* Credentials = 0
			);

		//! Get OBJITEM for raster.
		ERRVALUE GetObjItem (
			RVC::OBJITEM& ObjItem
			) const { return (m_Rast.GetObjItem(ObjItem)); }

		//! Open existing raster to draw into.
		ERRVALUE Open (
			const RVC::OBJITEM& ObjItem,
			const RVC::CREDENTIALS* Credentials = 0
			);

		//! Write the pixel array back to the raster.
		//!
		//! If the raster is small (less than 4096x4096), we just allocate a
		//! buffer big enough for the whole raster.  In this case, nothing
		//! gets flushed until we call WritePixArray() (which the destructor
		//! will do if necessary)
		ERRVALUE WritePixArray (
			);
			
	protected:

		//! In the Raster case, this will actually read the raster
		virtual UINT8* v_GetImagePointer (
			INT32 x,
			INT32 y,
			INT32 len,
			UINT16 FrameBuffer
			);

		//! If GetImagePointer returns a pointer directly into the frame buffer,
		//!   this function does not need to be overridden.
		//! If GetImagePointer has to actually read something (say via an X
		//!   call to the display or reading out of a raster, then this version
		//!   should just return a temporary buffer big enough to hold a single
		//!   line.
		virtual UINT8* v_GetImagePointerNoRead (
			INT32 x,
			INT32 y,
			INT32 len,
			UINT16 FrameBuffer
			);

		//! Write the pixel array from raster to buffer
		//!
		//! If the raster is small (less than 4096x4096), we just allocate a
		//! buffer big enough for the whole raster.  
		ERRVALUE ReadPixArray (
			);

		
		//! Write an image buffer to the display
		//! If GetImagePointer reutrns a pointer directly into the frame buffer,
		//! this method can just return 0.  If GetImagePointer has to actually
		//! read something, this method is used to write the result back out.
		//! Typically used when drawing to a raster
		virtual ERRVALUE v_WriteImageBuffer (
			INT32 x,
			INT32 y,
			INT32 len,
			UINT8 *buf,
			UINT16 FrameBuffer
			);

	private:
   	#ifndef GENERATING_DOXYGEN_OUTPUT
		friend class CONTEXT_RASTER_RGB24;

		//! Copy constructor private and unimplemented
		DEVICE_RASTER_RGB24(const DEVICE_RASTER_RGB24&);

		RVC::RASTER m_Rast;

		bool m_bUseMemory;
		bool m_bDirty;			//!< need to flush
		int m_TileSizeX;
		int m_TileSizeY;

		SIMPLE_ARRAY<UINT8> m_buffer;
		#endif // GENERATING_DOXYGEN_OUTPUT

		//! Create a GC for drawing.  
		//!
		//! Derived classes may override this and make it return anything
		//! derived from MGD_GC.  MGD_GC's constructor is private, so this
		//! is the only way for calling code to create a GC.
		virtual MGD::CONTEXT* v_CreateContext ();

	};

class CONTEXT_RASTER_RGB24 : public CONTEXT_MEM24 {
	public:
		// Note, default Constructor is protected

		virtual ~CONTEXT_RASTER_RGB24();

	protected:
		//! Default Constructor 
		CONTEXT_RASTER_RGB24 (
			DEVICE_RASTER_RGB24* dev
			);

		DEVICE_RASTER_RGB24* m_DevRast;

	private:
   	#ifndef GENERATING_DOXYGEN_OUTPUT

		//! Copy constructor private and unimplemented
		CONTEXT_RASTER_RGB24(const CONTEXT_RASTER_RGB24&);

		friend class MGD::DEVICE_RASTER_RGB24;	// Only DEVICE_RASTER_RGB24::CreateContext() can call our constructor
	
		#endif // GENERATING_DOXYGEN_OUTPUT

	};


//=====================================================================================================================

//! Device for rendering to 32-bit RVC raster.
//! Raster must have CELLTYPE_RedGreenBlueAlpha_32 or AlphaRedGreenBlue_32.
class DEVICE_RASTER_RGB32 : public DEVICE_MEM32RGBA {

	public:

		//! Construct a 32-bit RGB raster device.
		//! After constructing, you should call Create()
		DEVICE_RASTER_RGB32();

		//! Destructor
		virtual ~DEVICE_RASTER_RGB32();

		// (Note: copy constructor private and unimplemented.  Can't copy one.)

		//! Create a raster to draw into.
		ERRVALUE Create (
			RVC::OBJECT& ParentObject,
			const RVC::DESCRIPTOR& Descriptor,
			RVC::RASTER::MAKEPARMS& MakeParms,
			const RVC::CREDENTIALS* Credentials = 0
			);

		//! Create a raster to draw into.
		ERRVALUE Create (
			RVC::OBJITEM& ObjItem,
			RVC::RASTER::MAKEPARMS& MakeParms,
			const RVC::CREDENTIALS* Credentials = 0
			);

		//! Get OBJITEM for raster.
		ERRVALUE GetObjItem (
			RVC::OBJITEM& ObjItem
			) const { return (m_Rast.GetObjItem(ObjItem)); }

		//! Open existing raster to draw into.
		ERRVALUE Open (
			const RVC::OBJITEM& ObjItem,
			const RVC::CREDENTIALS* Credentials = 0
			);

		//! Write the pixel array back to the raster.
		//!
		//! If the raster is small (less than 4096x4096), we just allocate a
		//! buffer big enough for the whole raster.  In this case, nothing
		//! gets flushed until we call WritePixArray() (which the destructor
		//! will do if necessary)
		ERRVALUE WritePixArray ();
			
	protected:

		class CONTEXT : public DEVICE_MEM32RGBA::CONTEXT {
			public:

				virtual ~CONTEXT ();

			protected:

				CONTEXT (
					DEVICE_RASTER_RGB32 *pDevice
					);

			private:
		   	#ifndef GENERATING_DOXYGEN_OUTPUT
				CONTEXT (const CONTEXT&);

				friend class MGD::DEVICE_RASTER_RGB32;
				#endif // GENERATING_DOXYGEN_OUTPUT

			};

		//! In the Raster case, this will actually read the raster
		virtual UINT8* v_GetImagePointer (
			INT32 x,
			INT32 y,
			INT32 len,
			UINT16 FrameBuffer
			);

		//! If GetImagePointer returns a pointer directly into the frame buffer,
		//!   this function does not need to be overridden.
		//! If GetImagePointer has to actually read something (say via an X
		//!   call to the display or reading out of a raster, then this version
		//!   should just return a temporary buffer big enough to hold a single
		//!   line.
		virtual UINT8* v_GetImagePointerNoRead (
			INT32 x,
			INT32 y,
			INT32 len,
			UINT16 FrameBuffer
			);

		//! Write the pixel array from raster to buffer
		//!
		//! If the raster is small (less than 4096x4096), we just allocate a
		//! buffer big enough for the whole raster.  
		ERRVALUE ReadPixArray (
			);

		//! Write an image buffer to the display
		//! If GetImagePointer reutrns a pointer directly into the frame buffer,
		//! this method can just return 0.  If GetImagePointer has to actually
		//! read something, this method is used to write the result back out.
		//! Typically used when drawing to a raster
		virtual ERRVALUE v_WriteImageBuffer (
			INT32 x,
			INT32 y,
			INT32 len,
			UINT8 *buf,
			UINT16 FrameBuffer
			);

	private:
   	#ifndef GENERATING_DOXYGEN_OUTPUT

		//! Copy constructor private and unimplemented
		DEVICE_RASTER_RGB32(const DEVICE_RASTER_RGB32&);

		RVC::RASTER m_Rast;

		bool m_bUseMemory;
		bool m_bDirty;			//!< need to flush
		int m_TileSizeX;
		int m_TileSizeY;

		SIMPLE_ARRAY<UINT8> m_buffer;

		virtual MGD::CONTEXT* v_CreateContext ();
		#endif // GENERATING_DOXYGEN_OUTPUT

	};


//=====================================================================================================================

class CONTEXT_RASTER_RGB16;

//! Device for rendering to 16-bit RVC rasters
class DEVICE_RASTER_RGB16: public DEVICE_MEM16 {

	public:

		//! Construct 16-bit raster device
		//! After constructing, you should call Create()
		DEVICE_RASTER_RGB16();

		//! Destructor
		virtual ~DEVICE_RASTER_RGB16();

		//! Create a raster to draw into.
		ERRVALUE Create (
			RVC::OBJITEM& ObjItem,
			RVC::RASTER::MAKEPARMS& MakeParms,
			const RVC::CREDENTIALS* Credentials = 0
			);

		//! Open existing raster to draw into.
		ERRVALUE Open (
			const RVC::OBJITEM& ObjItem,
			const RVC::CREDENTIALS* Credentials = 0
			);

		//! Write the pixel array back to the raster.
		//!
		//! If the raster is small (less than 4096x4096), we just allocate a
		//! buffer big enough for the whole raster.  In this case, nothing
		//! gets flushed until we call WritePixArray() (which the destructor
		//! will do if necessary)
		ERRVALUE WritePixArray (
			);

	protected:

		//! Write the pixel array from raster to buffer
		//!
		//! If the raster is small (less than 4096x4096), we just allocate a
		//! buffer big enough for the whole raster.  
		ERRVALUE ReadPixArray (
			);

		
	private:
   	#ifndef GENERATING_DOXYGEN_OUTPUT
		friend class CONTEXT_RASTER_RGB16;

		//! Copy constructor private and unimplemented
		DEVICE_RASTER_RGB16(const DEVICE_RASTER_RGB16&);

		RVC::RASTER m_Rast;
		bool m_bUseMemory;
		bool m_bDirty;			//!< need to flush
		int m_TileSizeX;
		int m_TileSizeY;

		SIMPLE_ARRAY<UINT8> m_buffer;

		//! Create a GC for drawing.  
		//!
		//! Derived classes may override this and make it return anything
		//! derived from MGD_GC.  MGD_GC's constructor is private, so this
		//! is the only way for calling code to create a GC.
		virtual CONTEXT* v_CreateContext (
			);

		//! In the Raster case, this will actually read the raster
		virtual UINT8* v_GetImagePointer (
			INT32 x,
			INT32 y,
			INT32 len,
			UINT16 FrameBuffer
			);

		//! If GetImagePointer returns a pointer directly into the frame buffer,
		//!   this function does not need to be overridden.
		//! If GetImagePointer has to actually read something (say via an X
		//!   call to the display or reading out of a raster, then this version
		//!   should just return a temporary buffer big enough to hold a single
		//!   line.
		virtual UINT8* v_GetImagePointerNoRead (
			INT32 x,
			INT32 y,
			INT32 len,
			UINT16 FrameBuffer
			);

		//! Write an image buffer to the display
		//! If GetImagePointer reutrns a pointer directly into the frame buffer,
		//! this method can just return 0.  If GetImagePointer has to actually
		//! read something, this method is used to write the result back out.
		//! Typically used when drawing to a raster
		virtual ERRVALUE v_WriteImageBuffer (
			INT32 x,
			INT32 y,
			INT32 len,
			UINT8* buf,
			UINT16 FrameBuffer
			);
		#endif // GENERATING_DOXYGEN_OUTPUT

	};


class CONTEXT_RASTER_RGB16 : public CONTEXT_MEM16 {
	public:
		// Note, default Constructor is protected

		virtual ~CONTEXT_RASTER_RGB16();


	protected:
		//! Default Constructor 
		CONTEXT_RASTER_RGB16 (
			DEVICE_RASTER_RGB16* dev
			);

		DEVICE_RASTER_RGB16* m_DevRast;

	private:
   	#ifndef GENERATING_DOXYGEN_OUTPUT

		//! Copy constructor private and unimplemented
		CONTEXT_RASTER_RGB16(const CONTEXT_RASTER_RGB16&);

		#endif // GENERATING_DOXYGEN_OUTPUT
		friend class MGD::DEVICE_RASTER_RGB16;	// Only DEVICE_RASTER_RGB24::CreateContext() can call our constructor
	
	};


//=====================================================================================================================

} // End of namespace MGD


#endif	// INC_MI32_MG2RAST_H

