/******************************************************************************
 *
 * \file mi32/mg2simpl.h
 * \brief A base implementation of MGD::DEVICE and MGD::CONTEXT
 *
 * \if NODOC
 * $Log: mg2simpl.h_v $
 * Revision 1.22  2004/02/17 18:04:38  dwilliss
 * Oops.  Can't make the FillSpan optional in both versions or it's ambiguious
 *
 * Revision 1.21  2004/02/17 18:02:49  dwilliss
 * FillSpan is now optional
 *
 * Revision 1.20  2004/02/17 17:53:36  dwilliss
 * Made it possible to construct with a callback that takes CONTEXT_SIMPLE*
 * instead of just CONTEXT*.  Why make the callback cast it?
 *
 * Revision 1.19  2004/02/13 22:01:22  scowan
 * Removed old device name defines.
 *
 * Revision 1.18  2003/10/01 22:16:48  dwilliss
 * Doxygen
 *
 * Revision 1.17  2003/09/15 13:49:56  fileserver!dwilliss
 * Doxygen
 *
 * Revision 1.16  2003/06/11 17:12:42  dwilliss
 * Virtual methods all renamed to start with v_ and made private (where
 * possible) or protected.
 *
 * Revision 1.15  2002/08/19 15:17:37  scowan
 * Fixed g++ warning.
 * Formatting adjustments.
 *
 * Revision 1.14  2002/01/08 18:57:25  dwilliss
 * Removed a layer of classes
 *
 * Revision 1.13  2002/01/04 20:51:07  dwilliss
 * Rename a method for consistancy
 *
 * Revision 1.12  2002/01/04 18:15:12  dwilliss
 * DeviceGeneric now has GetSetFillSpanFunc methods and a CONTEXT_GENERIC
 * which has a GetPrivData method
 *
 * Revision 1.11  2002/01/03 22:06:18  dwilliss
 * Various Fixes
 *
 * Revision 1.10  2002/01/02 22:49:44  dwilliss
 * Everything compiles now
 *
 * Revision 1.9  2001/12/31 16:57:34  dwilliss
 * GetColorSpace no longer pure virtual.
 *
 * Revision 1.8  2001/12/26 20:24:47  dwilliss
 * *** empty log message ***
 *
 * Revision 1.7  2001/12/26 18:25:14  dwilliss
 * Cast return type to avoid stupid compiler warning
 *
 * Revision 1.6  2001/12/24 21:59:44  dwilliss
 * *** empty log message ***
 *
 * Revision 1.5  2001/12/24 16:23:08  dwilliss
 * Added NoMultiPointSpan support
 *
 * Revision 1.4  2001/08/31 16:42:21  dwilliss
 * Get/SetOutputScale()
 *
 * Revision 1.3  2001/07/16 14:56:23  dwilliss
 * Fixed some more mac errors
 *
 * Revision 1.2  2001/07/13 15:30:02  dwilliss
 * Some Mac compile issues
 *
 * Revision 1.1  2001/07/11 17:33:56  dwilliss
 * Initial revision
 *
 * \endif
 */
#ifndef INC_MI32_MG2SIMPL_H
#define INC_MI32_MG2SIMPL_H

#ifndef INC_MI32_MGD2_H
#include <mi32/mgd2.h>
#endif


namespace MGD {

class CONTEXT_SIMPLE;

//! This class is an MGD::DEVICE that lets you pass a pointer to a
//! fill span function.  This is mostly for use by old code that used to
//! use MgdInitSimple().  In fact, MgdInitSimple() is now implemented
//! using this class.
class DEVICE_SIMPLE : public DEVICE {
	public:
		typedef int (*OLDFILLSPANFUNC)(MGD::CONTEXT* gc, INT32 x1, INT32 x2, INT32 y);
		typedef int (*FILLSPANFUNC)(MGD::CONTEXT_SIMPLE* gc, INT32 x1, INT32 x2, INT32 y);

		DEVICE_SIMPLE (
			INT32 height,
			INT32 width,
			INT32 depth,
			OLDFILLSPANFUNC FillSpan,
			void* PrivData = 0
			);

		DEVICE_SIMPLE (
			INT32 height,
			INT32 width,
			INT32 depth,
			FILLSPANFUNC FillSpan = 0,
			void* PrivData = 0
			);

		// Note: constructor is protected.  Can only construct a derived class

		//! Destructor
		virtual ~DEVICE_SIMPLE();

		// (Note: copy constructor private and unimplemented.  Can't copy one.)

		void* GetPrivData (
			) const { return (m_PrivData); }

		ERRVALUE SetDisplayInfo (
			const DISPLAYINFO& dispinfo
			) { return SetDeviceInfo(dispinfo); }

		FILLSPANFUNC GetFillSpanFunc(
			) const { return (m_FillSpan); }

		void SetFillSpanFunc (
			FILLSPANFUNC FillSpanFunc
			) { m_FillSpan = FillSpanFunc; }

		void SetFillSpanFunc (
			OLDFILLSPANFUNC FillSpanFunc
			) { m_FillSpan = (FILLSPANFUNC)FillSpanFunc; }

		void SetPrivData (
			void* data
			) { m_PrivData = data; }

	private:
   	#ifndef GENERATING_DOXYGEN_OUTPUT
		COLORSPACE m_ColorSpace;
		FILLSPANFUNC m_FillSpan;
		void* m_PrivData;

		DEVICE_SIMPLE(const DEVICE_SIMPLE&);
		#endif // GENERATING_DOXYGEN_OUTPUT

		CONTEXT* v_CreateContext();

		//! Return the "native" COLORSPACE used by the display
		//! Image operations using this colorspace will be more effeciant.
		virtual COLORSPACE v_GetColorSpace (
			) const { return (m_ColorSpace); }

		virtual ERRVALUE v_FillSpan (
			MGD::CONTEXT* gc,
			INT32 x1,
			INT32 x2,
			INT32 y
			) {
			// Can't use static_cast here because CONTEXT_SIMPLE isn't defined
			// yet, so static_cast doesn't know that it's derived from CONTEXT
			return (m_FillSpan((MGD::CONTEXT_SIMPLE*)gc, x1, x2, y));
			}

	};



class CONTEXT_SIMPLE : public CONTEXT {
	public:
		CONTEXT_SIMPLE (
			DEVICE* dev
			) : CONTEXT(dev)
			{
			}
	
		void* GetPrivData() const {
			return ((static_cast<DEVICE_SIMPLE&>(GetDevice())).GetPrivData());
			}
	};

} // End of namespace MGD


#endif	// INC_MI32_MG2SIMPL_H

