/**
 * \file mi32/socket.h
 * \brief A Netowrk Socket class
 *
 * \if NODOC
 * $Log: socket.h_v $
 * Revision 1.23  2007/04/05 16:48:00  dwilliss
 * Rearrange nested classes so hopefully doxygen will behave
 *
 * Revision 1.22  2007/03/30 19:52:43  dwilliss
 * don't hide nested class defines from Doxygen
 *
 * Revision 1.21  2006/11/30 18:58:46  scowan
 * Added methods on ipaddr and a ipaddr list class.
 *
 * Revision 1.20  2006/11/30 17:18:19  scowan
 * Added isRoutable to ipaddr.
 *
 * Revision 1.19  2006/06/21 16:12:35  dwilliss
 * another attempt to get doxygen to behave
 *
 * Revision 1.18  2006/06/20 15:44:10  dwilliss
 * trying to get Doxygen to behave
 *
 * Revision 1.17  2006/04/12 21:29:43  scowan
 * *** empty log message ***
 *
 * Revision 1.16  2006/04/11 14:48:21  dwilliss
 * doxygen problem
 *
 * Revision 1.15  2006/03/22 23:44:45  dwilliss
 * Added GetSockName
 *
 * Revision 1.14  2005/08/29 20:08:21  scowan
 * Forward declare mi string to fix g++ warnings.
 *
 * Revision 1.13  2005/07/20 14:28:43  mju
 * Include stdio hdr for platform-specific FILE typedef (debug mode only).
 *
 * Revision 1.12  2005/07/14 17:54:29  mju
 * Fix file typedef to match std C.
 *
 * Revision 1.11  2005/07/14 17:49:45  mju
 * Forward-declare file struct if in debug mode.
 *
 * Revision 1.10  2005/04/27 15:38:38  dwilliss
 * Added Get/SetLocal to ADDR
 *
 * Revision 1.9  2005/04/12 14:47:41  dwilliss
 * Renamed our class to avoid name conflict with Microsoft include files
 *
 * Revision 1.8  2005/04/11 16:26:29  dwilliss
 * try to fix doxygen output
 *
 * Revision 1.7  2005/02/24 16:47:31  dwilliss
 * Added an IsConnected method to MISOCKET::CLIENT
 *
 * Revision 1.6  2004/03/12 15:46:08  dwilliss
 * Gave MISOCKET::CLIENT a LOCK class
 *
 * Revision 1.5  2004/03/11 21:19:40  dwilliss
 * Changed access of socket client virtual methods
 * cope with destroying client while it's still in use
 *
 * Revision 1.4  2004/02/18 17:51:03  dwilliss
 * Added some debugging support for the debug build
 *
 * Revision 1.3  2004/01/30 20:49:11  dwilliss
 * changed v_Write to return ERRVALUE instead of INT32
 *
 * Revision 1.2  2004/01/30 18:01:58  dwilliss
 * Fixed problem that VC6 didn't catch but gcc did
 *
 * Revision 1.1  2004/01/02 22:57:52  dwilliss
 * Initial revision
 *
 * \endif
 **/

#ifndef INC_MI32_SOCKET_H
#define INC_MI32_SOCKET_H

#ifndef INC_MI32_GENERICRW_H
#include <mi32/genericrw.h>
#endif

#ifndef INC_MI32_MILIST_H
#include <mi32/milist.h>
#endif

#if defined(DEBUG) && !defined(INC_STDIO_H)
#include <stdio.h>	// For FILE typedef needed in debug mode
#define	INC_STDIO_H
#endif

// The rest of the world uses these but for Windows, they're defined in winsock.h
// and we don't want to force the programmer to include that mess.
// 	struct sockaddr;
// 	struct sockaddr_in;
// 	struct in_addr;



#ifdef MISYSTEMDLL
	#define CLASSLIBEXPORT MI_DLLCLASSEXPORT
#else
	#define CLASSLIBEXPORT MI_DLLCLASSIMPORT
#endif

#ifndef GENERATING_DOXYGEN_OUTPUT
class STRUTF8;
class MISTRING;
#endif

//! The base class for MISOCKET_TCP and MISOCKET_LOCAL
//! The base class does not have Open or Close methods.  
//! Bind and are implemented in the derived classes, as each 
//!   implementation may take different parameters.
//! You close a socket by deleting the class instance.
class CLASSLIBEXPORT MISOCKET : public GENERICRW {
	public:

		enum FAMILY {
			FAMILY_NONE = 0,		//!< Equivilant to AF_UNSPEC
			FAMILY_LOCAL = 1,		//!< Equivilant to AF_UNIX
			FAMILY_INET = 2		//!< Equivilant to AF_INET
			};

		enum TYPE {
			TYPE_NONE = 0,			//!< Invalid value for unitilized structures
			TYPE_STREAM = 1,		//!< Equivilant to SOCK_STREAM.  Provides sequenced, reliable, two-way connection-based byte streams.
			TYPE_DGRAM = 2 		//!< Equivilant to SOCK_DGRAM.  Connectionless, unreliable buffers of a fixed maximum length.
			};

		class IPADDRLIST;

		//! \struct MISOCKET::IPADDR
		//! An Internet Protocol address.
		//! Note: this is equivilant to struct in_addr
		//! the member names are the same for compatibility reasons.
		struct CLASSLIBEXPORT IPADDR {
			union {	
				struct { UINT8 s_b1, s_b2, s_b3, s_b4; } S_un_b;
				UINT32 S_addr;	
				} S_un;

			//! Empty constructor
			IPADDR (
				); 

			//! Initialize with a known address.
			//! Address must be in network byte order (hi-lo) XXX ?maybe?
			IPADDR (
				UINT32 addr
				);

			//! Compare two IPADDR's.
			//! @return -1 if lhs < rhs, 0 if equal, 1 if lhs > rhs		
			int Compare (
				const MISOCKET::IPADDR& rhs
				) const;

			//! Get the list of available IP addresses for this machine		
			//! NOTE: Only implemented for Windows
			static void GetAddressList (
				IPADDRLIST& AddrList,			//!< List of IP addresses RETURNED
				bool OnlyRoutable = false		//!< Only list routable addresses
				);

			void GetString (
				MISTRING& mistr,
				bool bDoReverseDNSLookup
				) const;
				
			//! Is 'this' a lookback address according to RFC1918 and RFC3330 ?
			//! @return 'True' if the address is a loopback address (127.0.0.0 - 127.255.255.255)
			bool IsLoopbackAddr (
				) const;
			
			//! Is 'this' a routable address according to RFC1918 and RFC3330 ?
			//! Basically, the address is not in the 10 A class private range, the 192.168 B class private range
			//! the 172.16 - 172.31 B class private ranges, the 169.254 class B autoconfiguration range,
			//! the 127 class A loopback range, and the 224 - 239 class A multicast ranges
			//! @return 'True' if the address is not in the ranges listed above, 'false' it the address is.
			bool IsRoutable (
				) const;
			
			//! Set the address by looking up a host name.
			//! \return true if the name could be resolved, false if not.
			bool SetByName (
				const char* name
				);

			};		// MISOCKET::IPADDR


		//! \struct MISOCKET::ADDR
		//! A generic socket address.
		//! Note: this is equivilant to struct sockaddr_storage, which can be used
		//! to represent all sockaddr types.  
		//! Defining this ourselves allows us to use it without having to include
		//! winsock.h on Windows machines.
		//!
		//!	XXX THIS IS UGLY AND EVIL 
		//!
		struct CLASSLIBEXPORT ADDR {

			unsigned short sa_family;	// For compatibility with struct sockaddr
			char __ss_pad1[6];
			INT64 __ss_align;				// Assures correct alignment
			char __ss_pad2[112];

			//! Empty constructor
			ADDR (
				);
			
			//! Construct for an IP address
			ADDR (
				const MISOCKET::IPADDR& ip,
				int port
				);

			//! The the address in a humanly-readable format.
			//! \param mistr The string to place the address in
			//! \param bDoReverseDNSLookup If true, FAMILY_NET addresses will use 
			//! reverse DNS to find the hostname of the address.  If false, it will 
			//! just use the numeric IP address.  Ignored for other address families.
			void GetAddress (
				MISTRING& mistr,
				bool bDoReverseDNSLookup
				) const;

			//! Get the socket family
			MISOCKET::FAMILY GetFamily (
				) const;
			
			//! Get the IP Address (FAMILY_INET only).
			const MISOCKET::IPADDR& GetIPAddress (
				) const;

			//! Get the socket name (FAMILY_LOCAL only)
			const char* GetLocalName (
				) const;

			//! Get the port (FAMILY_INET only)
			int GetPort (
				) const;

			//! Construct for an IP address
			void Set (
				const MISOCKET::IPADDR& ip,
				int port
				);

			//! Set the socket family
			void SetFamily (
				MISOCKET::FAMILY family
				);

			//! Set the socket name (FAMILY_LOCAL only)
			void SetLocalName (
				const char* name
				);

			protected:
				const char* GetBytePtr (int offset) const;
				char* GetBytePtr (int offset);
				UINT16 GetUINT16(int offset) const;
				UINT32 GetUINT32(int offset) const;
				void SetUINT16(int offset, UINT16 val);
				void SetUINT32(int offset, UINT32 val);

			}; // MISOCKET::ADDR




		//! \class MISOCKET::CLIENT
		//! Base class for any client that wants to be notified of a socket's state.
		class CLASSLIBEXPORT CLIENT {
			public:
				CLIENT();

				//! Constructor
				//! \param sock The socket to watch. Can be NULL to indicate that we're 
				//! not watching anything.
				//! \param bTakeOwnership If true (the default) then this CLIENT
				//! "owns" the socket and will delete it.
				CLIENT (
					MISOCKET* sock, 
					bool bTakeOwnership = true
					);

				//! Destructor.
				//! If we "own" a socket, it will be deleted as well.
				//! Warning:  Do NOT destroy client from within one if its virtual 
				//!	methods.  The caller of that virtual method may not be done
				//!   with the "this" yet.  Call Destroy() from withing virtual methods.
				virtual ~CLIENT();

				bool CallOnAccept (
					MISOCKET* socket,
					const MISOCKET::ADDR& addr
					);

				//! Will be called when the socket is closed out from under us 
				//! (dropped connection or some such event).
				//! WARNING:  Do NOT destroy "this" from within this method or
				//! bad things may happen.  Call Destroy() instead.
				void CallOnClose (
					);

				//! Will be called when a Connect call completes if the socket was
				//! non blocking.
				void CallOnConnect (
					);

				//! Will be called when there is data available if a previous read 
				//! couldn't be filled and the socket was non-blocking.
				void CallOnReadReady (
					);

				//! Will be called if a timeout interval has been set and a previous
				//! call has timed out.
				void CallOnTimeOut (
					);
				
				//! Will be called when the socket is ready to be written to
				void CallOnWriteReady (
					);

				//! Connect to a remote server
				ERRVALUE Connect (
					const MISOCKET::ADDR& address
					);

				//! Get the current socket for this client.
				//! \return a MISOCKET or NULL if no socket being monitored.
				MISOCKET* GetSocket (
					) const;


				//! Determine if this client is connected to anything
				bool IsConnected (
					) const;

				//! Set the socket to watch.
				//! If we're already watching a socket, that socket is first informed
				//! that we're no longer watching it.  If we "own" the existing socket,
				//! it is deleted.
				//! If the socket is deleted out from under us, its destructor will
				//! reset our socket back to NULL.
				//! \param sock The socket to watch. Can be NULL to indicate that we're 
				//! not watching anything.
				//! \param bTakeOwnership If true (the default) then this CLIENT
				//! "owns" the socket and will delete it.
				void SetSocket (
					MISOCKET* sock,
					bool bTakeOwnership = true
					);
				
			protected:

				// A resource locking class which will prevent Destroy() from actually
				// deleting the client.
				class LOCK {
					public:
						LOCK (
							CLIENT* client
							) :
							m_client(client) 
							{
							if (m_client != 0) m_client->Lock();
							}
						~LOCK (
							) {
							if (m_client != 0) m_client->Unlock();
							}
					private:
						CLIENT* m_client;
					};

				//! Destroy the client.  
				//! For use from within virtual methods.
				void Destroy (
					);

			private:

				//! Will be called if a previous call to Accept() failed because
				//! the socket is non-blocking and the call would have blocked, but
				//! would now succeed.
				//! Will be passed a pointer to the already accepted socket. 
				//! If you wish to reject the connection, return false, and the caller
				//!   will delete the socket for you.
				//! If you wish to accept the connection, return true.  You are then
				//!	responsible for deleting the socket when you're done with it.
				virtual bool v_OnAccept (
					MISOCKET* socket,
					const MISOCKET::ADDR& addr
					);

				//! Will be called when the socket is closed out from under us 
				//! (dropped connection or some such event).
				//! WARNING:  Do NOT destroy "this" from within this method or
				//! bad things may happen.  Call Destroy() instead.
				virtual void v_OnClose (
					);

				//! Will be called when a Connect call completes if the socket was
				//! non blocking.
				virtual void v_OnConnect (
					);

				//! Will be called when there is data available if a previous read 
				//! couldn't be filled and the socket was non-blocking.
				virtual void v_OnReadReady (
					);

				//! Will be called if a timeout interval has been set and a previous
				//! call has timed out.
				virtual void v_OnTimeOut (
					);
				
				//! Will be called when the socket is ready to be written to
				virtual void v_OnWriteReady (
					);
#ifndef GENERATING_DOXYGEN_OUTPUT
			private:
				CLIENT(const CLIENT&);	// private and unimplemented
				CLIENT& operator=(const CLIENT&);
				friend class LOCK;		// Only MISOCKET::CLIENT::LOCK can call Lock/Unlock

				MISOCKET* m_sock;
				bool m_bWeOwnSocket;
				bool m_bNeedToDestroy;
				int m_LockCount;

				//! Prevent Destroy() from actually deleting this
				//! This is a ref-count lock.
				void Lock();

				//! Unlock "this" so that Destroy() can actually delete this.
				//! If Destroy() was called while locked, unlocking will delete this.
				void Unlock();
#endif	// GENERATING_DOXYGEN_OUTPUT
			};


		//! Default constructor
		MISOCKET (
			);

		//! Destructor.
		//! If this socket has a CLIENT bound to it, the client will be
		//! reset to a NULL socket.
		virtual ~MISOCKET (
			);


		//! Accept an incoming connection.
		//! The socket must be in a listening state. 
		//! If the socket is non-blocking and has a CLIENT attached to 
		//! it, then this function will always return 0 and the client's
		//! OnAccept() will be called when an incoming connection is available
		//! Creates and returns a new socket for the connection.  
		//! The address returned tells you where the incoming connection is from.
		//! It is up to the caller to delete the MISOCKET returned when done.
		MISOCKET* Accept (
			ADDR& address			//!< Address of incoming connection (returned)
			);

		//! Bind the socket to a given address
		//! This is done if you plan to make the socket into a listening socket
		//! or if you want all outgoing connections ton this socket to come from
		//! as specific address or port
		ERRVALUE Bind (
			const ADDR& address
			);

		//! Connect to a given server
		//! If the socket is non-blocking, Connect may return before the 
		//! connection is actually made. 
		//! In this case, a CLIENT attached to the socket will have its
		//! OnConnect() called when the connection completes or OnTimeOut if it
		//! doesn't complete.
		ERRVALUE Connect (
			const ADDR& address
			);

		//! Returns the MISOCKET::ADDR of the socket.
		//! If not bound to any address, the returned address will be
		//! unitialized and will have a family of FAMILY_NONE
		const ADDR& GetAddress (
			) const;

		//! Returns the address of the socket in a humanly-readable form
		//! \param mistr The string to place the address in
		//! \param bDoReverseDNSLookup If true, IP sockets will use reverse DNS
		//! to find the hostname of the address.  If false, it will just return
		//! the numeric IP address.  Ignored for non-IP sockets.
		void GetAddress (
			MISTRING& mistr,
			bool bDoReverseDNSLookup = false
			) const;
		
		//! Gets the address of local end of a connected socket.
		//! This is useful on a multi-homed server to find out which interface
		//!	you're connected through.
		//! \return true if the address was filled in, false if not.
		bool GetSockName (
			ADDR& addr
			) const;

		//! Get the timeout interval in seconds
		int GetTimeOut (
			) const;
			
		//! Determine if the socket is nonblocking
		bool IsNonBlocking (
			) const;

		//! Put the socket into listening mode.
		//! Must be previously Bound
		ERRVALUE Listen (
			);

		//! Read a single new-line terminated string (stripping out the newline).
		//! If there is no newline in the currently available input buffer this
		//! function will either wait for one or return false if the socket is
		//! non-blocking.
		//! Note: This function assumes that the incoming data is utf-8
		//! @return 1 if a string was read, 0 if the socket is non-blocking
		//! and no string available yet, < 0 on error.
		ERRVALUE ReadStr (
			STRUTF8& string
			);

		//! Set the timeout for I/O operations.
		//! To disable timeouts, set it to 0.
		void SetTimeOut (
			int seconds		// XXX Make this double seconds or integer ms
			);

		//! Set the socket to non-blocking mode.
		//! In non-blocking mode, if a read request cannot be fulfilled with
		//! what's currently available, it will return 0 and not read anything.
		void SetNonBlocking (
			bool bNonBlocking = true
			);
		
		//! Allocates a new socket of a given family and type.
		ERRVALUE Socket (
			FAMILY family,			//!< Socket family
			TYPE type = TYPE_STREAM		//!< Socket type (default is STREAM)
			);

		//! Write a NULL-terminated string.
		ERRVALUE WriteStr (
			const char* str
			);

#ifndef GENERATING_DOXYGEN_OUTPUT
		// These are public so that implementations can derive from them
		// but they're not exported from misystem
		class FACTORY;
		class IMPL;

#ifdef DEBUG
		// These three are only in the debug build
		void DumpDebugState (
			);
		void SetDebugFile (
			FILE*
			);
		void SetCloseDebug (
			);
#endif
#endif

	private:
#ifndef GENERATING_DOXYGEN_OUTPUT
		IMPL* m_impl;

		// Copy constructor and assignment operators are private - do not use
		MISOCKET(const MISOCKET&);
		MISOCKET& operator=(const MISOCKET&);
		// For internal use.
		MISOCKET(IMPL*);

		friend class CLIENT;	// I only want CLIENT to be able to call SetClient()
		friend class IMPL;	// I only want CLIENT to be able to call SetClient()

		//! Set the client to notify when data is ready.
		//! This would be implemented using the subject/observer pattern, but
		//! a socket can really only have one observer.  So implementing it this
		//! way not only simplifies the implementation, it enforces this restriction.
		//! Client can be null to stop notifications.
		//! Note: You cannot call this method yourself.  It is called for you by
		//! CLIENT::SetSocket()
		void SetClient (
			CLIENT* client
			);
		
		
		// Methods overridden from the base class (GENERICRW)
		
		GENERICRW::CAPABILITIES v_GetCapabilities (
			) const;

		// For this implementation, return the ammount of data available to read
		// right now.
		INT64 v_GetSize (
			) const;

		INT64 v_GetSizeLimit (
			) const;

		// Note: Read will block if AllowPartial is false and the requested
		// ammount of data is not available, unless the socket is marked
		// NonBlocking.  If it's marked non-blocking and there is not enough
		// data available, Read will return 0.
		INT32 v_Read (
			void* buffer,
			INT32 numbytes,
			bool AllowPartial = false
			);

		ERRVALUE v_Resize (
			INT64 newsize
			);

		INT64 v_Seek (
			INT64 offet,
			SEEKFROM from
			);

		INT64 v_Tell (
			) const;

		ERRVALUE v_Write (
			const void* buffer,
			INT32 numbytes
			);


#endif	// GENERATING_DOXYGEN_OUTPUT
	};

//! The inlines below are for the support of MILIST<MISOCKET::IPADDR>
inline bool operator== (
	const MISOCKET::IPADDR& lhs,
	const MISOCKET::IPADDR& rhs
	) { return (lhs.Compare(rhs) == 0); }

inline bool operator!= (
	const MISOCKET::IPADDR& lhs,
	const MISOCKET::IPADDR& rhs
	) { return (lhs.Compare(rhs) != 0); }

inline bool operator< (
	const MISOCKET::IPADDR& lhs,
	const MISOCKET::IPADDR& rhs
	) { return (lhs.Compare(rhs) == -1); }

class CLASSLIBEXPORT MISOCKET::IPADDRLIST : public MILIST<MISOCKET::IPADDR> {};

#undef CLASSLIBEXPORT

#endif // INC_MI32_SOCKET_H
// vim: ts=3 sw=3



