/**
 * \file rvc/geotable.h 
 * \brief interface for the RVC::SHAPE class.
 *
 * \if NODOC
 * $Id: geotable.h_v 1.5 2012/10/09 15:00:41 mju Exp $
 *
 * $Log: geotable.h_v $
 * Revision 1.5  2012/10/09 15:00:41  mju
 * Add "separate" to setupElemTransList.
 *
 * Revision 1.4  2008/09/04 13:10:09  mju
 * Remove unused getRecordNum method.
 *
 * Revision 1.3  2004/07/12 23:31:56  scowan
 * Removed use of geo region.
 *
 * Revision 1.1  2004/03/25 20:55:33  scowan
 * Initial revision
 *
 * \endif
**/

//!	\class RVC::GEOTABLE rvc/geotable.h
//!	\brief Manages RVC::OBJTYPE_GeoTable (Pinmap like) RVC objects.
//!	
//!	RVC Object Types that this class supports: RVC::OBJTYPE_GeoTable.
//!
//!	Default RVC Object Type for this class: RVC::OBJTYPE_GeoTable.
//!
//!	Valid parents for this class:
//!		- RVC::OBJTYPE_Database
//!		- RVC::OBJTYPE_Database_Raster
//!		- RVC::OBJTYPE_Database_VectorPoint
//!		- RVC::OBJTYPE_Database_VectorLine
//!		- RVC::OBJTYPE_Database_VectorPolygon
//!		- RVC::OBJTYPE_Database_CAD
//!		- RVC::OBJTYPE_Database_TIN
//!		- RVC::OBJTYPE_Database_Shape

#ifndef INC_RVC_GEOTABLE_H
#define INC_RVC_GEOTABLE_H

#ifndef INC_RVC_GEOMETRC_H
	#include <rvc/geometrc.h>
#endif

#ifndef INC_RVC_DBTABLE_H
	#include <rvc/dbtable.h>
#endif

#ifndef GENERATING_DOXYGEN_OUTPUT
class QUADTREE;
#endif // GENERATING_DOXYGEN_OUTPUT

namespace RVC {

class GEOTABLE : public GEOMETRIC {
	public:

		#ifndef GENERATING_DOXYGEN_OUTPUT
		class ELEMITERATOR;
		friend class ELEMITERATOR;
		class MAKEPARMS;
		#endif // GENERATING_DOXYGEN_OUTPUT
		
		//! Default constructor
		GEOTABLE (
			);

		//! Copy constructor
		GEOTABLE (
			const RVC::GEOTABLE& rhs
			);

		//! Destructor
		virtual ~GEOTABLE (
			);

		//! Assignment, will close and re-open object if possible
		RVC::GEOTABLE& operator= (
			const RVC::GEOTABLE& rhs
			);

		//! Obtain an element iterator to the beginning of the element list
		ELEMITERATOR ElementBegin (
			) const;

		//! Obtain an element iterator to the end of the element list
		ELEMITERATOR ElementEnd (
			) const;
			
		//! Are the fields in the database for the points editable (can they be changed)
		bool IsEditable (
			) const;
			
		//! Create new shape object given database table, creation parameters, and source
		ERRVALUE Make (
			const RVC::DBTABLE& table,
			const MAKEPARMS& MakeParms,
			const char* Source = 0
			);
		
		//! Create new shape object given RVC::OBJITEM, creation parameters, and source
		ERRVALUE Make (
			const RVC::OBJITEM& ObjItem,
			const MAKEPARMS& MakeParms,
			const char* Source = 0
			);

		//! Read a point from the GeoTable object source
		ERRVALUE Read (
			INT32 RecNum,
			DPOINT3D& Point
			) const;
			
		//! Read a record from the GeoTable object source
		ERRVALUE Read (
			INT32 RecNum,
			DBTABLE::RECORD& Record
			) const;
			
		//! Modify an existing point from the GeoTable object source
		//! GEOTABLE::IsEditable() must return 'true' for this to work.
		ERRVALUE Write (
			INT32 RecNum,
			const DPOINT3D& Point
			);
			
	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		//! Implementation of RVC::OBJECT virtual methods
		virtual ERRVALUE v_CloseObject ();
		virtual OBJTYPE v_GetDftObjectType () const;
		virtual bool v_IsObjectOpen () const;
		virtual ERRVALUE v_OpenObject (OPENMODE OpenFlags, MDLGPARENT parent);

		//! Implementation of RVC::GEOSPATIAL virtual methods
		virtual void v_GetExtents (DRECT3D &extents) const;
		virtual void v_GetScale (double& xscale, double& yscale) const;
		virtual double v_GetZOffset () const;
		virtual double v_GetZScale () const;
		virtual bool v_HasCapability (int Capability) const;
		virtual void v_SetScale (double xscale, double yscale);
		virtual void v_SetZOffset (double zoffset);
		virtual void v_SetZScale (double zscale);

		//! Implementation of RVC::GEOMETRIC virtual methods
		virtual ERRVALUE v_AddSearchTreeElement (INT32 elemnum, ELEMTYPE ElemType, const DRECT2D& Extents);
		virtual ERRVALUE v_BuildSearchTree (INT32 InitialNumElements, ELEMTYPE ElemType, const DRECT2D& Extents);
		virtual ERRVALUE v_ClearSearchTree (INT32 InitialNumElements, ELEMTYPE ElemType, const DRECT2D& Extents);
		virtual ERRVALUE v_CopySearchTreeElement (INT32 SourceElement, INT32 DestElement, ELEMTYPE ElemType, const DRECT2D& Extents);
		virtual ERRVALUE v_DeleteSearchTreeElement (INT32 elemnum, ELEMTYPE ElemType);
		virtual ERRVALUE v_ExtractTo (RVC::GEOMETRIC& DestObj, RVC::GEOMETRIC::ELEMTRANSLIST& TransList, const TRANS2D_MAPGEN& tmg, RVC::GEOMETRIC::EXTRACTFLAGS ExtractFlags) const;
		virtual INT32 v_FindClosestElementLow (ELEMTYPE ElemType, const DPOINT2D& SearchPt, GEOMETRIC::SEARCHFILTER* SearchFilter, double SearchDist, double* DistFromElem) const;
		virtual ERRVALUE v_FindClosestElementsLow (const DPOINT2D& SearchPt, double SearchDist, SEARCHFILTER* SearchFilter, SIMPLE_ARRAY<RVC::ELEMENT>& ElementList) const;
		virtual ERRVALUE v_FindElementDrawSet (ELEMTYPE ElemType, const DRECT2D& SearchExtents, SEARCHRESULT& SearchResult) const;
		virtual ERRVALUE v_FindElementSearchSet (ELEMTYPE ElemType, const DRECT2D& SearchExtents, SEARCHRESULT& SearchResult) const;
		virtual ERRVALUE v_GetElementExtents (ELEMTYPE ElemType, INT32 ElemNum, DRECT3D& Extents) const;
		virtual INT32 v_GetNextElementID (ELEMTYPE ElemType) const;
		virtual INT32 v_GetNumElements (ELEMTYPE ElemType) const;
		virtual POINTTYPE v_GetPointType () const;
		virtual INT32 v_GetPointSize () const;
		virtual bool v_IsElementDeleted (const RVC::ELEMENT& Element) const;
		virtual bool v_IsEmpty () const;
		virtual ERRVALUE v_SaveSearchTrees ();
		virtual ERRVALUE v_SetExtents (const DRECT3D &extents);
		virtual ERRVALUE v_SetNextElementID (ELEMTYPE ElemType, INT32 NextElementID);
		virtual bool v_SetSearchTree (ELEMTYPE ElemType, RVC::QSEARCHTREE* SearchTree, bool ReplaceExisting);
		virtual int v_SetupElemTransList (const GEOMETRIC::ELEMSELECTPARMS& SelectParms, GEOMETRIC::ELEMTRANSLIST& TransList, GEOMETRIC::ELEMSELECTFILTER *SelectFilter, bool separate) const;
		virtual bool v_TestElementAgainstRegion (const RVC::ELEMENT& Element, const REGION2D& region, TESTCRITERIA TestMode) const;
		virtual ERRVALUE v_UpdateExtents (const DRECT3D &extents);

		ERRVALUE ComputePoint (INT32 RecNum, DPOINT3D& point);
				
		DBTABLE m_DBTableObj;
		INT32 m_XFieldNum;
		INT32 m_YFieldNum;
		INT32 m_ZFieldNum;
		MISTRING m_GeoScript;
		SMLCONTEXT* m_Context;
		SMLSEGMENT m_Segment;
		class PRIV;
		mutable PRIV* m_Priv;
		mutable QUADTREE* m_QTree;
		mutable bool m_SetInstanceData;		//!< This object set the instance data if true, it needs to remove it on closing
		#endif // GENERATING_DOXYGEN_OUTPUT
	};
	

//! Constant iterator class
//!
//! This class is used to iterate through a geotable object elements
//! See methods ElementBegin() and ElementEnd() for obtaining an element iterator
class GEOTABLE::ELEMITERATOR {
	public:

		//! Default constructor
		ELEMITERATOR (
			) : 
			m_CurrentObj(0),
			m_IsDone(false)
			{}

		//! Internal constructor used by ElementEnd() methods
		ELEMITERATOR (
			const RVC::GEOTABLE* GeoTableObj,
			bool IsDone
			) : 
			m_CurrentObj(GeoTableObj)
			m_IsDone(IsDone)
			{}

		//! Copy constructor
		ELEMITERATOR (
			const ELEMITERATOR& rhs
			) : 
			m_CurrentObj(rhs.m_CurrentObj),
			m_IsDone(rhs.m_IsDone),
			m_RecIt(rhs.m_RecIt)
			{}

		//! Internal constructor used by ElementBegin() methods
		ELEMITERATOR (
			const RVC::GEOTABLE* GeoTableObj
			) :
			m_CurrentObj(GeoTableObj),
			m_IsDone(false)
			{
			GeoTableObj->RecordBegin(m_RecIt);
			}

		//! Destructor
		virtual ~ELEMITERATOR (
			) {}

		//! Assignment
		ELEMITERATOR& operator= (
			const ELEMITERATOR& rhs
			) {
			if (this != &rhs) {
				m_CurrentObj = rhs.m_CurrentObj;
				m_IsDone = rhs.m_IsDone;
				m_RecIt = rhs.m_RecIt;
				}
			return (*this);
			}

		//! Dereference
		const DPOINT3D& operator* (
			) const;

		//! Dereference arrow
		const DPOINT3D* operator-> (
			) const { return (&**this); }

		//! Pre-increment
		ELEMITERATOR& operator++ (
			) {
			m_RecIt++;
			m_IsDone = m_RecIt.IsDone();
			return (*this);
			}			

		//! Comparison for equality
		bool operator== (
			const ELEMITERATOR& rhs
			) const { return (m_IsDone == rhs.m_IsDone && m_CurrentObj == rhs.m_CurrentObj); }

		//! Comparison for inequality
		bool operator!= (
			const ELEMITERATOR& rhs
			) const { return (!(*this == rhs)); }

		//! Get the current record 
		//! @return Record
		const DBTABLE::RECORD& GetRecord (
			) const { return (*m_RecIt); }

	protected:
		const GEOTABLE *m_CurrentObj;
		bool m_IsDone;
		DBTABLE::CONST_RECORDITERATOR m_RecIt;
		DPOINT3D m_Point;
	};
			

//! GeoTable object creation parameters
class GEOTABLE::MAKEPARMS : public RVC::GEOMETRIC::MAKEPARMSBASE {
	public:

		MAKEPARMS (
			INT32 XFieldNum,
			INT32 YFieldNum,
			INT32 ZFieldNum = -1
			) :
			GEOMETRIC::MAKEPARMSBASE((ZFieldNum == -1) ? POINTTYPE_2DXY : POINTTYPE_3DXY),
			m_XFieldNum(XFieldNum),
			m_YFieldNum(YFieldNum),
			m_ZFieldNum(ZFieldNum)
			{}

		MAKEPARMS (
			const OBJECTNAME& XFieldName,
			const OBJECTNAME& YFieldName
			) :
			GEOMETRIC::MAKEPARMSBASE(POINTTYPE_2DXY),
			m_XFieldNum(-1),
			m_YFieldNum(-1),
			m_ZFieldNum(-1),
			m_XFieldName(XFieldName),
			m_YFieldName(YFieldName)
			{}

		MAKEPARMS (
			const MISTRING& GeoScript
			) :
			GEOMETRIC::MAKEPARMSBASE(POINTTYPE_2DXY),
			m_XFieldNum(-1),
			m_YFieldNum(-1),
			m_ZFieldNum(-1),
			m_GeoScript(GeoScript)
			{}
			
		void SetZFieldNum (
			INT32 ZFieldNum
			) { 
			m_ZFieldNum = ZFieldNum; 
			SetPointType((ZFieldNum == -1) ? POINTTYPE_2DXY : POINTTYPE_3DXY);
			return;
			}

		void SetZFieldName (
			const OBJECTNAME& ZFieldName
			) { 
			m_ZFieldName = ZFieldName;
			SetPointType((ZFieldName.IsEmpty()) ? POINTTYPE_2DXY : POINTTYPE_3DXY);
			return;
			}

	private:
		INT32 m_XFieldNum;
		INT32 m_YFieldNum;
		INT32 m_ZFieldNum;
		OBJECTNAME m_XFieldName;
		OBJECTNAME m_YFieldName;
		OBJECTNAME m_ZFieldName;
		MISTRING m_GeoScript;

		friend class GEOTABLE;
	};
	
}		//! End of RVC namespace

#endif 	//!< INC_RVC_GEOTABLE_H
