/**
 * \file rvc/stylepatternline.h 
 * \brief interface for the RVC::STYLEPATTERN_LINE class.
 *
 * \if NODOC
 * $Id: stylepatternline.h_v 1.6 2007/04/04 17:49:29 scowan Exp $
 *
 * $Log: stylepatternline.h_v $
 * Revision 1.6  2007/04/04 17:49:29  scowan
 * Added virtual methods.
 *
 * Revision 1.5  2006/09/21 14:20:03  scowan
 * Added pattern type to the base.
 *
 * Revision 1.4  2006/09/20 20:01:11  dwilliss
 * We didn't have a GetFilled() method
 *
 * Revision 1.3  2006/09/20 17:27:57  dwilliss
 * nothing
 *
 * Revision 1.2  2006/09/20 16:53:52  dwilliss
 * Added ctor from raw buffer
 *
 * Revision 1.1  2006/09/18 17:54:35  scowan
 * Initial revision
 *
 * \endif
**/

#ifndef INC_RVC_STYLEPATTERNLINE_H
#define INC_RVC_STYLEPATTERNLINE_H

#ifndef INC_RVC_STYLEPATTERNBASE_H
	#include <rvc/stylepatternbase.h>
#endif

#ifndef INC_MI32_MG2ENUMS_H
	#include <mi32/mg2enums.h>
#endif

namespace RVC {

//! Line pattern class.	
class STYLEPATTERN_LINE : public STYLEPATTERN_BASE {
	public:

		enum ELEMENT_TYPE {
			ELEMENT_TYPE_Line =			0,
			ELEMENT_TYPE_CrossLine =	1,
			ELEMENT_TYPE_Circle =		2,
			ELEMENT_TYPE_Polyline =		3,
			ELEMENT_TYPE_Polygon =		4,
			ELEMENT_TYPE_Text =			5,
			ELEMENT_TYPE_Symbol =		6,
			ELEMENT_TYPE_COUNT
			};

		class ITERATOR;
		class ELEMENT;

		//! Pattern element referencing shared memory buffer.
		//! Primarily used by ITERATOR.
		class ELEMENT_SHARED {
			public:

				//! Get cap style.
				//! @return Line cap style.
				MGD::LINECAP GetCapStyle (
					) const { return ((m_pData->flags & LINEPATTELEM::FLAG_CapRound) ? MGD::LINECAP_Round : MGD::LINECAP_Butt); }

				//! Get color index.
				//! @return Index into color array.
				UINT8 GetColorIndex (
					) const { return (m_pData->colornum); }

				bool GetFilled (
					) const { return ((m_pData->flags & LINEPATTELEM::FLAG_Filled) != 0); }

				//! Get interval between repeated elements.
				//! @return Interval in 1/100 millimeter.
				INT16 GetInterval (
					) const { return (m_pData->xspacing); }
					
				MGD::LINEJOIN GetJoinStyle (
					) const { return ((m_pData->flags & LINEPATTELEM::FLAG_JoinMiter) ? MGD::LINEJOIN_Miter : (m_pData->flags & LINEPATTELEM::FLAG_JoinBevel) ? MGD::LINEJOIN_Bevel : MGD::LINEJOIN_Round); }

				//! Get offset to start of element along line from beginning.
				//! @return Offset in 1/100 millimeter.
				INT16 GetOffsetAlong (
					) const { return (m_pData->xoffset); }

				//! Get offset aside line (left -, right +).
				//! @return Offset in 1/100 millimeter.
				INT16 GetOffsetAside (
					) const { return ((m_pData->type != ELEMENT_TYPE_Line) ? -m_pData->yoffset : m_pData->yoffset); }

				//! Get size/length of element.
				//! @return Size in 1/100 millimeter.
				INT16 GetSize (
					) const { return (m_pData->xsize); }

				//! Get element line thickness.
				//! @return Thickness in 1/100 millimeter.
				INT16 GetThickness (
					) const { return (m_pData->thickness); }

				//! Get element type.
				//! @return Element type.
				ELEMENT_TYPE GetType (
					) const { return (static_cast<ELEMENT_TYPE>(m_pData->type)); }

				//! Determine if element is 'filled'.
				//! @return 'true' if filled, 'false' if not.
				bool IsFilled (
					) const { return ((m_pData->flags & LINEPATTELEM::FLAG_Filled) != 0); }
					
			protected:

				struct LINEPATTELEM {
					enum TYPE {
						TYPE_Line =		0,				//!< Line parallel to drawing direction, xsize=0 indicates continuous
						TYPE_CrossLine,				//!< Line perpendicular to drawing direction
						TYPE_Circle,					//!< Circle - xsize = radius
						TYPE_PolyLine,					//!< PolyLine
						TYPE_Polygon,					//!< Polygon
						TYPE_Last = TYPE_Polygon	//!< Last defined type
						};
						
					enum FLAGS {
						FLAG_Filled =		0x0001,	//!< Fill (circle and polygon elements)
						FLAG_JoinMiter =	0x0002,	//!< Join "bent" segment mitered (extended)
						FLAG_JoinBevel =	0x0004,	//!< Join "bent" segment beveled
						FLAG_CapRound = 	0x0008	//!< Use rounded ends on line segments
						};

					INT16 extrabytes;							//!< Number of bytes in element not including header
					UINT8 type;									//!< Pattern element type (TYPE)
					UINT8 colornum;							//!< Color index in pattern-specific color map
					UINT16 flags;								//!< FLAGS
					INT16 xoffset;								//!< Offset along line from start
					INT16 xspacing;							//!< Spacing between instances of element
					INT16 yoffset;								//!< Offset above(-) or below(+) line
					INT16 xsize;								//!< Size/Length of element
					INT16 thickness;							//!< Thickness of element lines (if not filled)
					};


			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT

				const LINEPATTELEM *m_pData;
				void *m_pExtra;

				//! Default constructor.
				ELEMENT_SHARED (
					): m_pData(0), m_pExtra(0) { }

				explicit ELEMENT_SHARED (const void *pElemData)
					: m_pData(static_cast<const LINEPATTELEM*>(pElemData)), m_pExtra(static_cast<UINT8*>(const_cast<void*>(pElemData)) + sizeof(LINEPATTELEM))
					{ }

				//! Destructor.
				~ELEMENT_SHARED (
					) { }

				friend class RVC::STYLEPATTERN_LINE;
				friend class ELEMENT;
				friend class ITERATOR;
				#endif // GENERATING_DOXYGEN_OUTPUT
			};

		//! Pattern element.
		class ELEMENT : public ELEMENT_SHARED {
			public:

				//! Default constructor.
				ELEMENT (
					) { memset(&m_MyData,0,sizeof(m_MyData)); m_pData = &m_MyData; }

				//! Copy constructor.
				ELEMENT (
					const ELEMENT& rhs
					);

				//! Construct from ELEMENT_SHARED.
				ELEMENT (
					const ELEMENT_SHARED& rhs
					);

				//! Construct element of specified type.
				explicit ELEMENT (
					ELEMENT_TYPE type
					);

				//! Destructor.
				~ELEMENT (
					) { if (m_pExtra != 0) free(m_pExtra); }

				//! Assignment from ELEMENT.
				ELEMENT& operator= (
					const ELEMENT& rhs
					);

				//! Assignment from ELEMENT_SHARED.
				ELEMENT& operator= (
					const ELEMENT_SHARED& rhs
					);

				//! Set line 'cap style' for dashed and crossing lines.
				void SetCapStyle (
					MGD::LINECAP linecap
					) { if (linecap == MGD::LINECAP_Round) m_MyData.flags |= LINEPATTELEM::FLAG_CapRound; else m_MyData.flags &= ~LINEPATTELEM::FLAG_CapRound; }

				//! Set element color index.
				void SetColorIndex (
					UINT8 idx
					) { m_MyData.colornum = idx; }

				void SetFilled (
					bool filled
					) { if (filled) m_MyData.flags |= LINEPATTELEM::FLAG_Filled; else m_MyData.flags &= ~LINEPATTELEM::FLAG_Filled; }

				void SetInterval (
					INT16 interval
					) { m_MyData.xspacing = interval; }

				void SetJoinStyle (
					MGD::LINEJOIN linejoin
					) {
					m_MyData.flags &= ~(LINEPATTELEM::FLAG_JoinMiter|LINEPATTELEM::FLAG_JoinBevel);
					if (linejoin == MGD::LINEJOIN_Miter) m_MyData.flags |= LINEPATTELEM::FLAG_JoinMiter;
					else if (linejoin == MGD::LINEJOIN_Bevel) m_MyData.flags |= LINEPATTELEM::FLAG_JoinBevel;
					}

				void SetOffsetAlong (
					INT16 offset
					) { m_MyData.xoffset = offset; }

				void SetOffsetAside (
					INT16 offset
					) { m_MyData.yoffset = (m_MyData.type != ELEMENT_TYPE_Line) ? -offset : offset; }

				void SetSize (
					INT16 size
					) { m_MyData.xsize = size; }

				void SetThickness (
					INT16 thickness
					) { m_MyData.thickness = thickness; }

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				ELEMENT_SHARED::LINEPATTELEM m_MyData;
				#endif // GENERATING_DOXYGEN_OUTPUT
			};

		//! Iterator to process pattern elements.
		class ITERATOR {
			public:

				//! Default constructor.
				ITERATOR (
					): m_pEnd(0) { }

				//! Destructor.
				~ITERATOR (
					) { }

				//! Dereference operator, returns reference to ELEMENT_SHARED.
				const ELEMENT_SHARED& operator* (
					) const { return (m_element); }

				//! Dereference operator, returns pointer to ELEMENT_SHARED.
				const ELEMENT_SHARED* operator-> (
					) const { return (&m_element); }

				//! Pre-increment operator.
				ITERATOR& operator++ ();

				//! Equality operator
				bool operator== (
					const ITERATOR& rhs
					) const { return ((m_pEnd == 0 && rhs.m_pEnd == 0) || m_element.m_pData == rhs.m_element.m_pData); }

				//! Inequality operator
				bool operator!= (
					const ITERATOR& rhs
					) const { return (!(*this == rhs)); }

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT

				ELEMENT_SHARED m_element;
				const void *m_pEnd;

				ITERATOR (const void *pData, const void *pEnd)
					: m_element(pData), m_pEnd(pEnd)
					{ }

				friend class RVC::STYLEPATTERN_LINE;
				#endif // GENERATING_DOXYGEN_OUTPUT
			};

		//! Default constructor.
		STYLEPATTERN_LINE (
			) : STYLEPATTERN_BASE(PATTERNTYPE_Line, sizeof(PATTERNHEADER_LINE)) {}

		//! Constructor from raw pattern data
		STYLEPATTERN_LINE (
			const void* buf,
			INT32 NumBytes
			) : STYLEPATTERN_BASE(PATTERNTYPE_Line, sizeof(PATTERNHEADER_LINE)) {
			SetPattern(buf, NumBytes, sizeof(PATTERNHEADER_LINE));
			}
		
		//! Copy constructor.
		STYLEPATTERN_LINE (
			const STYLEPATTERN_LINE& rhs
			) : STYLEPATTERN_BASE(rhs) {}

		//! Destructor.
		virtual ~STYLEPATTERN_LINE (
			) { }

		//! Assignment.
		STYLEPATTERN_LINE& operator= (
			const STYLEPATTERN_LINE& rhs
			) { if (this != &rhs) { STYLEPATTERN_BASE::operator=(rhs); } return (*this); }

		//! Get iterator for beginning element.
		ITERATOR Begin (
			) const { return (ITERATOR(GetFirstElemPtr(),GetPattBuffer()+GetSize())); }

		//! Delete element at specified index.
		//! This will invalidate all ITERATORs.
		ERRVALUE DeleteElement (
			unsigned idx
			);

		//! Delete unused colors from pattern palette.
		//! This will invalidate all ITERATORs.
		void DeleteUnusedColors (
			);

		//! Get iterator for end comparison.
		ITERATOR End (
		#ifdef WIN32
			) const { return (ITERATOR()); }
		#else
			) const { ITERATOR it; return (it); }
		#endif

		//! Get element at specified index.
		//! Note, this is less efficient than using iterator.
		ELEMENT GetElement (
			unsigned idx
			) const;

		//! Get extent of pattern to 'left' as drawn (not including 'crossing' elements).
		INT16 GetExtentLeft (
			) const { return (reinterpret_cast<PATTERNHEADER_LINE*>(m_PattBuff)->extentsa); }

		//! Get extent of pattern to 'right' as drawn (not including 'crossing' elements).
		INT16 GetExtentRight (
			) const { return (reinterpret_cast<PATTERNHEADER_LINE*>(m_PattBuff)->extentsb); }

		//! Insert element at specified index.
		//! This will invalidate all ITERATORs.
		ERRVALUE InsertElement (
			unsigned idx,
			const ELEMENT& element
			);

		//! Set (replace) element at specified index.
		//! If element size changes this will invalidate all ITERATORs.
		ERRVALUE SetElement (
			unsigned idx,
			const ELEMENT& element
			);

		//! GetWidth - Return width if line pattern in 1/100th of a mm
		INT32 GetWidth (
			) const;

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		//! Line pattern header.
		struct PATTERNHEADER_LINE : public STYLEPATTERN_BASE::PATTERNHEADER {
			INT16 extentsa;							//!< Maximum extents of element "above" line (drawn left to right)
			INT16 extentsb;							//!< Maximum extents of element "below" line (drawn left to right)
			};

		ERRVALUE BuildOffsetIndex () const;
		INT32 GetOffset (unsigned idx) const;
		#endif // GENERATING_DOXYGEN_OUTPUT
	};

}		//! End of RVC namespace

#endif 	//!< INC_RVC_STYLEPATTERNLINE_H

