/**
 * \file soa/arcimsconnection.h
 * \brief SOA::ARCIMS::CONNECTION definition
 *
 * \if NODOC
 * $Id: arcimsconnection.h_v 1.5 2008/07/16 23:03:43 bcolombe Exp $
 *
 * $Log: arcimsconnection.h_v $
 * Revision 1.5  2008/07/16 23:03:43  bcolombe
 * Added BuildGetLegendRequest
 *
 * Revision 1.4  2007/09/28 23:05:06  dwilliss
 * Added image type
 *
 * Revision 1.3  2007/09/21 14:42:20  dwilliss
 * *** empty log message ***
 *
 * Revision 1.2  2007/09/14 21:45:54  dwilliss
 * It compiles now
 *
 * Revision 1.1  2007/09/14 16:57:40  dwilliss
 * Initial revision
 *
 *
 * \endif
**/

#ifndef	INC_SOA_ARCIMSCONNECTION_H
#define	INC_SOA_ARCIMSCONNECTION_H

#ifndef	INC_MI32_MISTRING_H
	#include <mi32/mistring.h>
#endif
#ifndef	INC_MI32_MIMETYPE_H
	#include <mi32/mimetype.h>
#endif

namespace SPATREF {
	class COORDREFSYS;
	}
namespace SPATMOD {
	namespace IMAGE {
		class SOURCE_HTML;
		}
	}
struct COLOR;
class MIMETYPE;
class HTTPMESSAGE;
class XMLDOC;

namespace SOA {
namespace ARCIMS {

class SERVICEINFO;

//=====================================================================================================================

//! Container for information about connection to a ARCIMS server.
//! Includes Server URL and capabilities, and provides methods to send service requests.
class CONNECTION {
	public:

		//! Return true if a URL is for this type of service
		static bool IsCompatibleURL (
			const MISTRING& url
			);

		//! Constructor.
		CONNECTION ();
		
		//! Copy constructor.
		CONNECTION (const CONNECTION&);

		//! Destructor.
		virtual ~CONNECTION ();

		//! Assignment.
		CONNECTION& operator= (const CONNECTION&);


		//! Build "GetFeatureInfo" request URL string
		ERRVALUE BuildGetFeatureInfoRequest (
			HTTPMESSAGE& RequestHdrs,								//!< Request headers to send
			XMLDOC& RequestBody,										//!< Request body to send
			const MISTRINGLIST& LayerNameList,					//!< List of layers to request
			const MISTRINGLIST& LayerStyleList,					//!< Corresponding layer style list
			const SPATREF::COORDREFSYS& CoordRefSys,			//!< CRS for bounding box and rendering
			const DRECT2D& BoundingBox,							//!< Bounding box
			INT32 TgtWidth,											//!< Target width in pixels
			INT32 TgtHeight,											//!< Target height in pixels
			const MIMETYPE& ImageFormat,							//!< Image format MIME type
			const MISTRINGLIST& QueryLayerList,					//!< List of layers to query
			const MIMETYPE& InfoFormat,							//!< Format for returning feature info
			INT32 PixelColumn,										//!< Column coordinate of position to query
			INT32 PixelRow,											//!< Row coordinate of position to query
			INT32 FeatureCount = 1									//!< Feature count
			) const;


		//! Build "GET_IMAGE" request message.
		//! Unlike WMS in which the request is an HTTP GET request, ArcIMS uses
		//! HTTP POST.  Also, the reply from a WMS GetMap request is the image.
		//! The reply from an ArcIMS GET_IMAGE is XML which may contain the image
		//! or a url to the image.
		ERRVALUE BuildGetImageRequest (
			HTTPMESSAGE& RequestHdrs,								//!< Request headers to send
			XMLDOC& RequestBody,										//!< Request body to send
			const MISTRINGLIST& LayerIDList,						//!< List of layers to request
			const SPATREF::COORDREFSYS& CoordRefSys,			//!< CRS for bounding box and rendering
			const DRECT2D& BoundingBox,							//!< Bounding box
			INT32 TgtWidth,											//!< Target width in pixels
			INT32 TgtHeight,											//!< Target height in pixels
			const MIMETYPE& Format,									//!< Format MIME type
			bool bRequestImage = true,								//!< Request main image
			const COLOR *pColorBG = 0								//!< Optional background color
			) const;

		ERRVALUE BuildGetLegendRequest (
			HTTPMESSAGE& RequestHdrs,								//!< Request headers to send
			XMLDOC& RequestBody,										//!< Request body to send
			const MISTRINGLIST& LayerIDList,						//!< List of layers to request
			const MIMETYPE& Format,									//!< Format MIME type
			const COLOR *pLegendBGColor = 0						//!< Optional background color for legend graphic
			) const;

		//! Send "GET_SERVICE_INFO" request to previously specified server.
		//! If already have capabilities will not request again.
		ERRVALUE DoGetServiceInfo (
			const MISTRING& ServerURL,
			bool synchronous,
			MDLGPARENT dlgparent = 0								//!< Parent for status dialog if want to show
			);

		//! Send "GET_SERVICE_INFO" request to previously specified server.
		//! If already have capabilities will not request again.
		ERRVALUE DoGetServiceInfo (
			bool synchronous,
			MDLGPARENT dlgparent = 0								//!< Parent for status dialog if want to show
			);

		const MIMETYPE& GetImageType (
			) const { return (m_ImageType); }

		const MISTRING& GetServerURL (
			) const { return (m_ServerURL); }

		//! Get capabilities as returned from server.
		//! If an asynchronous request for capabilities was made and which may still be open
		//! then must call HasCapabilities first and only call this method if returns true.
		const ARCIMS::SERVICEINFO& GetServiceInfo (
			) const;

		//! Get the title for the service.
		MISTRING GetServiceTitle (
			) const;

		//! Get the tooltip for the service.
		MISTRING GetServiceToolTip (
			) const;

		//! Determine if have service info available.
		bool HasServiceInfo (
			) const { return (m_bHasServiceInfo); }


		//! Set URL for server.
		//! If does not match existing then capabilities will be discarded.
		void SetServerURL (
			const MISTRING& ServerURL
			);

	private:
		#ifndef	GENERATING_DOXYGEN_OUTPUT
		MISTRING m_ServerURL;
		ARCIMS::SERVICEINFO *m_pServiceInfo;
		bool m_bHasServiceInfo;
		MIMETYPE m_ImageType;

		#endif
	};

//=====================================================================================================================

}	// End namespace ARCIMS
}	// End namespace SOA

#endif	// INC_SOA_ARCIMSCONNECTION_H
