/**
 * \file imagefilterconvertterrain.h <spatmod/imagefiltergeneral.h>
 * \brief Spatial Modelling (SPATMOD) IMAGE::FILTER_CONVERT_TERRAIN definitions.
 *
 * \if NODOC
 * $Id: imagefilterconvertterrain.h_v 1.6 2012/04/27 16:36:59 bcolombe Exp $
 *
 * $Log: imagefilterconvertterrain.h_v $
 * Revision 1.6  2012/04/27 16:36:59  bcolombe
 * Use scale numerator,denom
 *
 * Revision 1.5  2012/04/26 16:59:23  bcolombe
 * Remove precision, just use scale
 * Add methods for writing type,scale to tile
 *
 * Revision 1.4  2012/04/03 20:52:41  bcolombe
 * Changes for new terrain format
 *
 * Revision 1.3  2012/03/29 17:36:20  bcolombe
 * Add precision
 *
 * Revision 1.2  2012/03/15 16:06:37  bcolombe
 * Added scale to meters
 *
 * Revision 1.1  2012/03/14 22:00:45  bcolombe
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGEFILTERCONVERTTERRAIN_H
#define	INC_SPATMOD_IMAGEFILTERCONVERTTERRAIN_H

#ifndef	INC_SPATMOD_IMAGEFILTER_H
#include <spatmod/imagefilter.h>
#endif

#ifndef	INC_SPATMOD_IMAGETILESET_H
#include <spatmod/imagetileset.h>
#endif

namespace SPATMOD {
namespace IMAGE {

class BUFFER;
class SAMPLEITERATOR;

//=====================================================================================================================

//! Filter to convert between signed terrain and RGBA values
class FILTER_CONVERT_TERRAIN : public FILTER {
	public:
		// Constructor.
		FILTER_CONVERT_TERRAIN (
			STAGE *pPreviousStage,					//!< Pointer to previous stage
			bool TakeOwnership,						//!< Filter takes ownership of previous stage and will delete when needed
			const SPATMOD::IMAGE::TILESET::IMGFORMAT& ImgFormat, //!< ImgFormat of terrain
			bool bConvertTerrainToRGBA,						//!< True to convert to RGBA from signed terrain values, false to convert to terrain from RGBA
			double Scale = 1.0,
			INT64 PrescaleOffset = 0,
			INT64 PostscaleOffset = 0
			);

		FILTER_CONVERT_TERRAIN (
			STAGE *pPreviousStage,					//!< Pointer to previous stage
			bool TakeOwnership,						//!< Filter takes ownership of previous stage and will delete when needed
			const SPATMOD::IMAGE::FORMAT::SPEC& ImgFormat, //!< ImgFormat of terrain
			bool bConvertTerrainToRGBA,						//!< True to convert to RGBA from signed terrain values, false to convert to terrain from RGBA
			double Scale = 1.0,
			INT64 PrescaleOffset = 0,
			INT64 PostscaleOffset = 0
			);

		static void ReadPixelRange(
			const SPATMOD::IMAGE::FORMAT::SPEC& ImgFormat, 
			const BUFFER& Buffer, 
			double& MinValue, 
			double& MaxValue,
			double Scale = 1.0,
			INT64 PrescaleOffset = 0,
			INT64 PostscaleOffset = 0
			);

		static void ReadPixelRange(
			const SPATMOD::IMAGE::TILESET::IMGFORMAT& ImgFormat,
			const BUFFER& Buffer, 
			double& MinValue, 
			double& MaxValue,
			double Scale = 1.0,
			INT64 PrescaleOffset = 0,
			INT64 PostscaleOffset = 0
			);

		static void WritePixelRange(
			BUFFER& Buffer, 
			INT32 MinValue, 
			INT32 MaxValue
			);

		static void WriteScaleOffset(
			BUFFER& Buffer, 
			double Scale,
			INT64 PrescaleOffset = 0,
			INT64 PostscaleOffset = 0
			);

		static void WriteScaleOffset(
			BUFFER& Buffer, 
			INT32 ScaleNumerator,
			INT32 ScaleDenominator,
			INT64 PrescaleOffset = 0,
			INT64 PostscaleOffset = 0
			);

		static void WriteType(
			BUFFER& Buffer,
			INT32 Type
			);

		// Destructor.
		virtual ~FILTER_CONVERT_TERRAIN ();

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		SPATMOD::IMAGE::TILESET::IMGFORMAT m_ImgFormat;
		double m_Scale;
		INT64 m_PrescaleOffset;
		INT64 m_PostscaleOffset;
		bool m_bConvertTerrainToRGBA;

		bool ConvertRGBAToTerrain (
			const SPATMOD::IMAGE::SAMPLEITERATOR *ittgt,
			const SPATMOD::IMAGE::SAMPLEITERATOR *itsrc,
			bool valid
			);

		bool ConvertTerrainToRGBA (
			const SPATMOD::IMAGE::SAMPLEITERATOR *ittgt,
			const SPATMOD::IMAGE::SAMPLEITERATOR *itsrc,
			bool valid
			);

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_Initialize ();
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_CONVERT_TERRAIN (const FILTER_CONVERT_TERRAIN&);
		FILTER_CONVERT_TERRAIN& operator= (const FILTER_CONVERT_TERRAIN&);
		#endif // GENERATING_DOXYGEN_OUTPUT

	};
}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	// INC_SPATMOD_IMAGEFILTERCONVERTTERRAIN_H

