/**
 * \file imagesourcervc.h <spatmod/imagesourcervc.h>
 * \brief Spatial Modelling (SPATMOD) IMAGE::SOURCE_RVC definition.
 *
 * \if NODOC
 * $Id: imagesourcervc.h_v 1.19 2011/10/19 18:37:58 mju Exp $
 *
 * $Log: imagesourcervc.h_v $
 * Revision 1.19  2011/10/19 18:37:58  mju
 * Remove redundant hdrs.
 *
 * Revision 1.18  2011/03/31 15:26:21  mju
 * Add v_getTypeName.
 *
 * Revision 1.17  2010/01/26 18:03:20  mju
 * Rename member to m_useCacheRect.
 *
 * Revision 1.16  2009/03/10 14:08:28  mju
 * Change needTileSequential to flags.
 *
 * Revision 1.15  2008/05/05 14:02:36  mju
 * Add m_knowTierNeedSequential.
 *
 * Revision 1.14  2007/07/24 17:47:56  mju
 * Add setKeepOpen and supporting members.
 *
 * Revision 1.13  2007/07/09 20:55:24  mju
 * Add m_didSetCacheRect.
 *
 * Revision 1.12  2007/05/08 13:06:00  mju
 * Use objitemlist hdr.
 *
 * Revision 1.11  2007/04/09 13:43:24  mju
 * Remove dimension-access virtuals using zoom.
 *
 * Revision 1.10  2007/02/27 19:03:49  mju
 * Add v_getDimensions.
 *
 * Revision 1.9  2007/02/06 14:27:39  mju
 * Add setSpreadToByte.
 *
 * Revision 1.8  2007/01/03 23:03:04  mju
 * Add getMetadata.
 *
 * Revision 1.7  2006/11/30 15:37:04  mju
 * Add ctor with rvc credentials.
 *
 * Revision 1.6  2006/11/21 16:38:06  mju
 * Add setUsaValidity.
 *
 * Revision 1.5  2006/11/03 17:09:01  mju
 * Add optional sampleType parm to ctor.
 *
 * Revision 1.4  2006/11/01 23:10:27  mju
 * queryIterator.
 *
 * Revision 1.3  2006/01/25 17:34:26  mju
 * Add v_computeSampleStats.
 *
 * Revision 1.2  2005/08/19 14:26:35  mju
 * Remove setTotalSection.
 *
 * Revision 1.1  2005/08/16 15:10:59  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGESOURCERVC_H
#define	INC_SPATMOD_IMAGESOURCERVC_H

#ifndef	INC_SPATMOD_IMAGESOURCE_H
#include <spatmod/imagesource.h>
#endif

#ifndef	INC_SPATMOD_IMAGESAMPLE_H
#include <spatmod/imagesample.h>
#endif

#ifndef	INC_RVC_OBJITEMLIST_H
#include <rvc/objitemlist.h>
#endif

#ifndef	INC_RVC_RASTER_H
#include <rvc/raster.h>
#endif

#ifndef	INC_RVC_NULLMASK_H
#include <rvc/nullmask.h>
#endif

namespace SPATMOD {
namespace IMAGE {

//=====================================================================================================================

//! Image source from RVC raster object.
class SOURCE_RVC : public SOURCE {
	public:

		//! Standard constructor with optional component and sample type override.
		explicit SOURCE_RVC (
			const RVC::OBJITEM& ObjItem,					//!< Reference to object location
			RVC::RASTER::COMPONENT component = RVC::RASTER::COMPONENT_Default,	//!< Component, default uses all components
			SAMPLE::TYPE SampleType = SAMPLE::TYPE_Invalid							//!< Sample type override
			);

		//! Construct with RVC credentials and optional component and sample type override.
		SOURCE_RVC (
			const RVC::OBJITEM& ObjItem,					//!< Reference to object location
			const RVC::CREDENTIALS& Credentials,		//!< Credentials
			RVC::RASTER::COMPONENT component = RVC::RASTER::COMPONENT_Default,	//!< Component, default uses all components
			SAMPLE::TYPE SampleType = SAMPLE::TYPE_Invalid							//!< Sample type override
			);

		//! Destructor.
		virtual ~SOURCE_RVC ();

		//! Set whether to keep open after initialize outside pull begin/end processing.
		//! Must be done before Initialize().  Default is to keep open after initialize.
		void SetKeepOpen (
			bool KeepOpen
			) { m_KeepOpen = KeepOpen; }

		//! Set whether to spread < 8-bit values over byte range.
		//! Must be done before Initialize().  Default is to not spread.
		//! Has no effect for images having >= 8 bits per sample
		void SetSpreadToByte (
			bool SpreadToByte
			) { m_SpreadToByte = SpreadToByte; }

		//! Set whether to use color palette if available and appropriate for data type.
		//! Must be done before Initialize().  Default is to not use color palette.
		//! No processing of pixel values is done before using as index to palette, may
		//! only be used for unsigned data types.
		void SetUseColorPalette (
			bool UseColorPalette,							//!< True to use palette, false (default) to not
			const MIUNICODE* ObjNamePalette = 0			//!< Palette subobject name, default to last-used if NULL
			);

		//! Set whether to use pyramid tiers.
		//! Default is to use pyramid tiers for zooming if available.
		void SetUsePyramidTiers (
			bool UsePyramidTiers
			) { m_UsePyramidTiers = UsePyramidTiers; }

		//! Set whether to use validity (null mask/value) if available.
		//! Must be done before Initialize().  Default is to use validity.
		void SetUseValidity (
			bool UseValidity
			) { m_UseValidity = UseValidity; }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		SAMPLE::TYPE m_SampleType;
		RVC::RASTER::COMPONENT m_Component;
		RVC::IMAGE::CONVMODE m_ConvMode;
		RVC::CREDENTIALS *m_pCredentials;
		RVC::OBJITEM m_ObjItemMain;
		RVC::OBJITEMLIST m_ObjItemListTiers;
		SIMPLE_ARRAY<INT32> m_TierSampleSteps;
		mutable SIMPLE_ARRAY<UINT8> m_TierFlags;
		DIMENSIONS *m_TierDimensions;
		mutable RVC::RASTER m_ObjRaster;
		mutable RVC::NULLMASK m_ObjMask;
		mutable double m_CurZoom;
		bool m_UseColorPalette;
		bool m_UsePyramidTiers;
		bool m_UseValidity;
		bool m_SpreadToByte;
		bool m_KeepOpen;
		bool m_DoingPull;
		bool m_UseCacheRect;
		RVC::OBJECTNAME m_ObjNamePalette;
		RVC::RASTER::COMPONENT m_SampleComponent[4];
		SIMPLE_ARRAY<UINT16> m_SampleIdx;
		DOUBLE_ARRAY<double> m_RowBuf;

		static UINT8 CalcTierFlags (RVC::RASTER& ObjRaster);
		int GetTierIndex (double zoom) const;

		// BASE overrides.
		virtual ERRVALUE v_GetDimensions (DIMENSIONS& dimensions, double zoom) const;
		virtual float v_QueryIterator (const SECTION_ITERATOR& iterator) const;
		virtual const char* v_GetTypeName () const;
		virtual ZOOMABILITY v_GetZoomAbility (SIMPLE_ARRAY<double>& ZoomsAllowed) const;

		// STAGE overrides.
		virtual ERRVALUE v_ComputeSampleStats (STAT_UNIVARIATE *SampleStatsArray, STATSVALIDITY StatsValidity);
		virtual void v_GetMetadataThisOnly (MISTRING& result, METADATACONTENT content, METADATAFORMAT format) const;
		virtual ERRVALUE v_Initialize ();
		virtual ERRVALUE v_PullBegin (const SECTION_ITERATOR& iterator);
		virtual ERRVALUE v_PullEnd ();
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		SOURCE_RVC (const SOURCE_RVC&);
		SOURCE_RVC& operator= (const SOURCE_RVC&);
		#endif // GENERATING_DOXYGEN_OUTPUT

	};

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	// INC_SPATMOD_IMAGESOURCERVC_H

