/**
 * \file spatmod/imagetargetrvc.h
 * \brief Spatial Modelling (SPATMOD) IMAGE::TARGET_RVC definition.
 *
 * \if NODOC
 * $Id: imagetargetrvc.h_v 1.13 2012/02/01 21:51:48 mju Exp $
 *
 * $Log: imagetargetrvc.h_v $
 * Revision 1.13  2012/02/01 21:51:48  mju
 * Add output::isUsed.
 *
 * Revision 1.12  2011/10/19 18:40:52  mju
 * Remove redundant hdrs.
 *
 * Revision 1.11  2008/11/21 22:15:04  bcolombe
 * Added Close method to OUTPUT
 *
 * Revision 1.10  2007/09/20 14:28:15  mju
 * Add numExtraValidity.
 *
 * Revision 1.9  2007/07/10 17:58:50  mju
 * Add output.m_tilesToCache.
 *
 * Revision 1.8  2007/05/08 13:05:45  mju
 * Use objitemlist hdr.
 *
 * Revision 1.7  2006/12/13 21:12:54  mju
 * Remove computeCellTypeFromPixelProp.
 *
 * Revision 1.6  2006/11/30 15:37:21  mju
 * Support rvc credentials.
 *
 * Revision 1.5  2006/11/20 21:19:30  mju
 * Add computeCellTypeFromPixelProp.
 *
 * Revision 1.4  2006/11/01 23:10:39  mju
 * queryIterator.
 *
 * Revision 1.3  2006/07/28 13:32:40  mju
 * Use const in raster write.
 *
 * Revision 1.1  2005/08/16 15:11:13  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGETARGETRVC_H
#define	INC_SPATMOD_IMAGETARGETRVC_H

#ifndef	INC_SPATMOD_IMAGETARGET_H
#include <spatmod/imagetarget.h>
#endif

#ifndef	INC_RVC_OBJITEMLIST_H
#include <rvc/objitemlist.h>
#endif

#ifndef	INC_RVC_RASTER_H
#include <rvc/raster.h>
#endif

#ifndef	INC_RVC_NULLMASK_H
#include <rvc/nullmask.h>
#endif

namespace SPATMOD {
namespace IMAGE {

//=====================================================================================================================

//! RVC raster/image target.
class TARGET_RVC : public TARGET {
	public:

		//! Histogram modes.
		enum HISTOGRAM {
			HISTOGRAM_None,					//!< Don't compute histograms
			HISTOGRAM_Default,				//!< Default sample/full based on user preference setting
			HISTOGRAM_AutoSample,			//!< Auto-sample using pyramid tier if available
			HISTOGRAM_Full						//!< Compute full histogram
			};

		//! Construct target with single output object.
		//! If source image cannot be written to single object an error will occur when Initialize() is
		//! called, use constructor with OBJITEMLIST if multiple objects are to be created.
		TARGET_RVC (
			STAGE *pInputStage,								//!< Stage in pipeline to obtain image data from
			const RVC::OBJITEM& ObjItem,					//!< Object location and descriptor
			const RVC::CREDENTIALS *pCredentials = 0	//!< Optional credentials
			);

		//! Construct target with single output object.
		//! Number of items objects in list must be appropriate for input image or an error
		//! will be returned from Initialize().
		TARGET_RVC (
			STAGE *pInputStage,								//!< Stage in pipeline to obtain image data from
			const RVC::OBJITEMLIST& ObjItemList,		//!< Object locations and descriptors
			const RVC::CREDENTIALS *pCredentials = 0	//!< Optional credentials
			);

		//! Destructor.
		virtual ~TARGET_RVC ();

		//! Get object item list.
		//! Can be used after Initialize() to determine the actual locations of the objects created.
		const RVC::OBJITEMLIST& GetObjItemList (
			) const { return (m_ObjItemList); }

		//! Set compression type and quality.
		//! If used, must be called before Initialize().
		void SetCompression (
			RVC::RASTER::COMPTYPE type,
			UINT16 quality
			) { m_CompressionType = type; m_CompressionQuality = quality; }

		//! Set histogram creation mode.
		//! Histogram will be created at end of Process() only if entire extent of target image is specified.
		//! If used, must be called before Process().
		void SetHistogramMode (
			HISTOGRAM HistogramMode
			) { m_HistogramMode = HistogramMode; }

		//! Set pyramiding mode.
		//! Pyramiding will be done at end of Process() only if entire extent of target image is specified.
		//! If used, must be called before Process().
		//! By default pyramids are generated with user's preferred Full/Partial setting and pixel sampling.
		//! Using PYRAMID_Disable will also disable embedded tier creation for JPEG2000 compression.
		void SetPyramidMode (
			RVC::IMAGE::PYRAMID PyramidMode
			) { m_PyramidMode = PyramidMode; }

		//! Override default tile size.
		void SetTileSizeOverride (
			UINT32 ColumnsPerTile,					//!< Columsn per tile
			UINT32 RowsPerTile = 0					//!< Rows per tile, will use same as columns if 0
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		class OUTPUT {
			public:

				OUTPUT ();
				~OUTPUT ();

				void Close (
					) {
					m_ObjRaster.Close();
					m_ObjMask.Close();
					}

				RVC::RASTER& GetObjRaster (
					) { return (m_ObjRaster); }

				bool IsUsed () const
					{ return (m_ObjRaster.IsOpen()); }

				ERRVALUE Make (
					RVC::OBJITEM& ObjItem,
					RVC::IMAGE::CELLTYPE CellType,
					const TARGET_RVC& target
					);

				ERRVALUE SetCacheSize (
					const SECTION_ITERATOR& iterator
					);

				ERRVALUE Write (
					INT32 row,
					INT32 col,
					INT32 numcols,
					const void *data,
					const UINT8 *validity
					);

			private:
				RVC::RASTER m_ObjRaster;
				RVC::NULLMASK m_ObjMask;
				RVC::IMAGE::CONVMODE m_ConvMode;
				INT32 m_TilesToCache;
			};
		typedef MILIST<OUTPUT> OUTPUTLIST;
		friend class OUTPUT;

		RVC::OBJITEMLIST m_ObjItemList;
		RVC::CREDENTIALS *m_pCredentials;
		OUTPUTLIST m_OutputList;
		RVC::RASTER::COMPTYPE m_CompressionType;
		UINT16 m_CompressionQuality;
		HISTOGRAM m_HistogramMode;
		RVC::IMAGE::PYRAMID m_PyramidMode;
		UINT32 m_ColumnsPerTile;
		UINT32 m_RowsPerTile;
		SIMPLE_ARRAY<UINT16> m_SampleIdx;
		DOUBLE_ARRAY<double> m_RowBuf;
		bool m_NeedSequential;
		bool m_ProcessedFullImage;
		UINT16 m_NumExtraValidity;

		// BASE overrides.
		virtual float v_QueryIterator (const SECTION_ITERATOR& iterator) const;

		// TARGET overrides.
		virtual ERRVALUE v_Initialize ();
		virtual ERRVALUE v_PushBegin (const SECTION_ITERATOR& iterator);
		virtual ERRVALUE v_PushEnd ();
		virtual ERRVALUE v_PushSubSection (const SECTION& SubSection, const BUFFER& buffer);

		// Unimplemented
		TARGET_RVC (const TARGET_RVC&);
		TARGET_RVC& operator= (const TARGET_RVC&);
		#endif // GENERATING_DOXYGEN_OUTPUT

	};

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	// INC_SPATMOD_IMAGETARGETRVC_H

